# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.
from __future__ import absolute_import, division, print_function

import string
import re

try:
    import urllib.parse as urlparse
except ImportError:
    from urlparse import urlparse

from pyparsing import stringStart, stringEnd, originalTextFor, ParseException
from pyparsing import ZeroOrMore, Word, Optional, Regex, Combine
from pyparsing import Literal as L  # noqa

from poetry.semver import parse_constraint

from .markers import MARKER_EXPR, Marker


LEGACY_REGEX = r"""
    (?P<operator>(==|!=|<=|>=|<|>))
    \s*
    (?P<version>
        [^,;\s)]* # Since this is a "legacy" specifier, and the version
                  # string can be just about anything, we match everything
                  # except for whitespace, a semi-colon for marker support,
                  # a closing paren since versions can be enclosed in
                  # them, and a comma since it's a version separator.
    )
    """


REGEX = r"""
            (?P<operator>(~=|==|!=|<=|>=|<|>|===))
            (?P<version>
                (?:
                    # The identity operators allow for an escape hatch that will
                    # do an exact string match of the version you wish to install.
                    # This will not be parsed by PEP 440 and we cannot determine
                    # any semantic meaning from it. This operator is discouraged
                    # but included entirely as an escape hatch.
                    (?<====)  # Only match for the identity operator
                    \s*
                    [^\s]*    # We just match everything, except for whitespace
                              # since we are only testing for strict identity.
                )
                |
                (?:
                    # The (non)equality operators allow for wild card and local
                    # versions to be specified so we have to define these two
                    # operators separately to enable that.
                    (?<===|!=)            # Only match for equals and not equals
    
                    \s*
                    v?
                    (?:[0-9]+!)?          # epoch
                    [0-9]+(?:\.[0-9]+)*   # release
                    (?:                   # pre release
                        [-_\.]?
                        (a|b|c|rc|alpha|beta|pre|preview)
                        [-_\.]?
                        [0-9]*
                    )?
                    (?:                   # post release
                        (?:-[0-9]+)|(?:[-_\.]?(post|rev|r)[-_\.]?[0-9]*)
                    )?
    
                    # You cannot use a wild card and a dev or local version
                    # together so group them with a | and make them optional.
                    (?:
                        (?:[-_\.]?dev[-_\.]?[0-9]*)?         # dev release
                        (?:\+[a-z0-9]+(?:[-_\.][a-z0-9]+)*)? # local
                        |
                        \.\*  # Wild card syntax of .*
                    )?
                )
                |
                (?:
                    # The compatible operator requires at least two digits in the
                    # release segment.
                    (?<=~=)               # Only match for the compatible operator
    
                    \s*
                    v?
                    (?:[0-9]+!)?          # epoch
                    [0-9]+(?:\.[0-9]+)+   # release  (We have a + instead of a *)
                    (?:                   # pre release
                        [-_\.]?
                        (a|b|c|rc|alpha|beta|pre|preview)
                        [-_\.]?
                        [0-9]*
                    )?
                    (?:                                   # post release
                        (?:-[0-9]+)|(?:[-_\.]?(post|rev|r)[-_\.]?[0-9]*)
                    )?
                    (?:[-_\.]?dev[-_\.]?[0-9]*)?          # dev release
                )
                |
                (?:
                    # All other operators only allow a sub set of what the
                    # (non)equality operators do. Specifically they do not allow
                    # local versions to be specified nor do they allow the prefix
                    # matching wild cards.
                    (?<!==|!=|~=)         # We have special cases for these
                                          # operators so we want to make sure they
                                          # don't match here.
    
                    \s*
                    v?
                    (?:[0-9]+!)?          # epoch
                    [0-9]+(?:\.[0-9]+)*   # release
                    (?:                   # pre release
                        [-_\.]?
                        (a|b|c|rc|alpha|beta|pre|preview)
                        [-_\.]?
                        [0-9]*
                    )?
                    (?:                                   # post release
                        (?:-[0-9]+)|(?:[-_\.]?(post|rev|r)[-_\.]?[0-9]*)
                    )?
                    (?:[-_\.]?dev[-_\.]?[0-9]*)?          # dev release
                )
            )
"""


class InvalidRequirement(ValueError):
    """
    An invalid requirement was found, users should refer to PEP 508.
    """


ALPHANUM = Word(string.ascii_letters + string.digits)

LBRACKET = L("[").suppress()
RBRACKET = L("]").suppress()
LPAREN = L("(").suppress()
RPAREN = L(")").suppress()
COMMA = L(",").suppress()
SEMICOLON = L(";").suppress()
AT = L("@").suppress()

PUNCTUATION = Word("-_.")
IDENTIFIER_END = ALPHANUM | (ZeroOrMore(PUNCTUATION) + ALPHANUM)
IDENTIFIER = Combine(ALPHANUM + ZeroOrMore(IDENTIFIER_END))

NAME = IDENTIFIER("name")
EXTRA = IDENTIFIER

URI = Regex(r"[^ ]+")("url")
URL = AT + URI

EXTRAS_LIST = EXTRA + ZeroOrMore(COMMA + EXTRA)
EXTRAS = (LBRACKET + Optional(EXTRAS_LIST) + RBRACKET)("extras")

VERSION_PEP440 = Regex(REGEX, re.VERBOSE | re.IGNORECASE)
VERSION_LEGACY = Regex(LEGACY_REGEX, re.VERBOSE | re.IGNORECASE)

VERSION_ONE = VERSION_PEP440 ^ VERSION_LEGACY
VERSION_MANY = Combine(
    VERSION_ONE + ZeroOrMore(COMMA + VERSION_ONE), joinString=",", adjacent=False
)("_raw_spec")
_VERSION_SPEC = Optional(((LPAREN + VERSION_MANY + RPAREN) | VERSION_MANY))
_VERSION_SPEC.setParseAction(lambda s, l, t: t._raw_spec or "")

VERSION_SPEC = originalTextFor(_VERSION_SPEC)("specifier")
VERSION_SPEC.setParseAction(lambda s, l, t: t[1])

MARKER_EXPR = originalTextFor(MARKER_EXPR())("marker")
MARKER_EXPR.setParseAction(
    lambda s, l, t: Marker(s[t._original_start : t._original_end])
)
MARKER_SEPERATOR = SEMICOLON
MARKER = MARKER_SEPERATOR + MARKER_EXPR

VERSION_AND_MARKER = VERSION_SPEC + Optional(MARKER)
URL_AND_MARKER = URL + Optional(MARKER)

NAMED_REQUIREMENT = NAME + Optional(EXTRAS) + (URL_AND_MARKER | VERSION_AND_MARKER)

REQUIREMENT = stringStart + NAMED_REQUIREMENT + stringEnd


class Requirement(object):
    """Parse a requirement.

    Parse a given requirement string into its parts, such as name, specifier,
    URL, and extras. Raises InvalidRequirement on a badly-formed requirement
    string.
    """

    def __init__(self, requirement_string):
        try:
            req = REQUIREMENT.parseString(requirement_string)
        except ParseException as e:
            raise InvalidRequirement(
                'Invalid requirement, parse error at "{0!r}"'.format(
                    requirement_string[e.loc : e.loc + 8]
                )
            )

        self.name = req.name
        if req.url:
            parsed_url = urlparse.urlparse(req.url)
            if not (parsed_url.scheme and parsed_url.netloc) or (
                not parsed_url.scheme and not parsed_url.netloc
            ):
                raise InvalidRequirement("Invalid URL given")
            self.url = req.url
        else:
            self.url = None

        self.extras = set(req.extras.asList() if req.extras else [])
        constraint = req.specifier
        if not constraint:
            constraint = "*"

        self.constraint = parse_constraint(constraint)
        self.pretty_constraint = constraint

        self.marker = req.marker if req.marker else None

    def __str__(self):
        parts = [self.name]

        if self.extras:
            parts.append("[{0}]".format(",".join(sorted(self.extras))))

        if self.pretty_constraint:
            parts.append(self.pretty_constraint)

        if self.url:
            parts.append("@ {0}".format(self.url))

        if self.marker:
            parts.append("; {0}".format(self.marker))

        return "".join(parts)

    def __repr__(self):
        return "<Requirement({0!r})>".format(str(self))



import os
import re

from poetry.version.requirements import Requirement

from .dependency import Dependency
from .directory_dependency import DirectoryDependency
from .file_dependency import FileDependency
from .locker import Locker
from .package import Package
from .project_package import ProjectPackage
from .utils.link import Link
from .utils.utils import convert_markers
from .utils.utils import group_markers
from .utils.utils import is_archive_file
from .utils.utils import is_installable_dir
from .utils.utils import is_url
from .utils.utils import path_to_url
from .utils.utils import strip_extras
from .vcs_dependency import VCSDependency


def dependency_from_pep_508(name):
    # Removing comments
    parts = name.split("#", 1)
    name = parts[0].strip()
    if len(parts) > 1:
        rest = parts[1]
        if ";" in rest:
            name += ";" + rest.split(";", 1)[1]

    req = Requirement(name)

    if req.marker:
        markers = convert_markers(req.marker.markers)
    else:
        markers = {}

    name = req.name
    path = os.path.normpath(os.path.abspath(name))
    link = None

    if is_url(name):
        link = Link(name)
    else:
        p, extras = strip_extras(path)
        if os.path.isdir(p) and (os.path.sep in name or name.startswith(".")):

            if not is_installable_dir(p):
                raise ValueError(
                    "Directory {!r} is not installable. File 'setup.py' "
                    "not found.".format(name)
                )
            link = Link(path_to_url(p))
        elif is_archive_file(p):
            link = Link(path_to_url(p))

    # it's a local file, dir, or url
    if link:
        # Handle relative file URLs
        if link.scheme == "file" and re.search(r"\.\./", link.url):
            link = Link(path_to_url(os.path.normpath(os.path.abspath(link.path))))
        # wheel file
        if link.is_wheel:
            m = re.match("^(?P<namever>(?P<name>.+?)-(?P<ver>\d.*?))", link.filename)
            if not m:
                raise ValueError("Invalid wheel name: {}".format(link.filename))

            name = m.group("name")
            version = m.group("ver")
            dep = Dependency(name, version)
        else:
            name = link.egg_fragment

            if link.scheme == "git":
                dep = VCSDependency(name, "git", link.url_without_fragment)
            else:
                dep = Dependency(name, "*")
    else:
        if req.pretty_constraint:
            constraint = req.constraint
        else:
            constraint = "*"

        dep = Dependency(name, constraint)

    if "extra" in markers:
        # If we have extras, the dependency is optional
        dep.deactivate()

        for or_ in markers["extra"]:
            for _, extra in or_:
                dep.extras.append(extra)

    if "python_version" in markers:
        ors = []
        for or_ in markers["python_version"]:
            ands = []
            for op, version in or_:
                # Expand python version
                if op == "==":
                    version = "~" + version
                    op = ""
                elif op == "!=":
                    version += ".*"
                elif op == "in":
                    versions = []
                    for v in re.split("[ ,]+", version):
                        split = v.split(".")
                        if len(split) in [1, 2]:
                            split.append("*")
                            op = ""
                        else:
                            op = "=="

                        versions.append(op + ".".join(split))

                    if versions:
                        ands.append(" || ".join(versions))

                    continue

                ands.append("{}{}".format(op, version))

            ors.append(" ".join(ands))

        dep.python_versions = " || ".join(ors)

    if "sys_platform" in markers:
        ors = []
        for or_ in markers["sys_platform"]:
            ands = []
            for op, platform in or_:
                if op == "==":
                    op = ""
                elif op == "in":
                    platforms = []
                    for v in re.split("[ ,]+", platform):
                        platforms.append(v)

                    if platforms:
                        ands.append(" || ".join(platforms))

                    continue

                ands.append("{}{}".format(op, platform))

            ors.append(" ".join(ands))

        dep.platform = " || ".join(ors)

    # Extras
    for extra in req.extras:
        dep.extras.append(extra)

    return dep



"""
This toml module is a port with changes and fixes
of [contoml](https://github.com/jumpscale7/python-consistent-toml).
"""


from .toml_file import TOMLFile
from .prettify.lexer import tokenize as lexer
from .prettify.parser import parse_tokens


def loads(text):
    """
    Parses TOML text into a dict-like object and returns it.
    """
    tokens = tuple(lexer(text, is_top_level=True))
    elements = parse_tokens(tokens)

    return TOMLFile(elements)


def load(file_path):
    """
    Parses a TOML file into a dict-like object and returns it.
    """
    with open(file_path) as fd:
        return loads(fd.read())


def dumps(value):
    """
    Dumps a data structure to TOML source code.

    The given value must be either a dict of dict values, a dict,
    or a TOML file constructed by this module.
    """
    if not isinstance(value, TOMLFile):
        raise RuntimeError(
            "Can only dump a TOMLFile instance loaded by load() or loads()"
        )

    return value.dumps()


def dump(obj, file_path, prettify=False):
    """
    Dumps a data structure to the filesystem as TOML.

    The given value must be either a dict of dict values, a dict,
    or a TOML file constructed by this module.
    """
    with open(file_path, "w") as fp:
        fp.write(dumps(obj))



from .venv_command import VenvCommand


class BuildCommand(VenvCommand):
    """
    Builds a package, as a tarball and a wheel by default.

    build
        { --f|format= : Limit the format to either wheel or sdist. }
    """

    def handle(self):
        from poetry.masonry import Builder

        fmt = "all"
        if self.option("format"):
            fmt = self.option("format")

        package = self.poetry.package
        self.line(
            "Building <info>{}</> (<comment>{}</>)".format(
                package.pretty_name, package.version
            )
        )

        builder = Builder(self.poetry, self.venv, self.output)
        builder.build(fmt)



import os

from ..command import Command


class CacheClearCommand(Command):
    """
    Clears poetry's cache.

    cache:clear
        { cache : The name of the cache to clear. }
        { --all : Clear all caches. }
    """

    def handle(self):
        from cachy import CacheManager
        from poetry.locations import CACHE_DIR

        cache = self.argument("cache")

        parts = cache.split(":")
        cache_dir = os.path.join(CACHE_DIR, "cache", "repositories", parts[0])
        cache = CacheManager(
            {
                "default": parts[0],
                "serializer": "json",
                "stores": {parts[0]: {"driver": "file", "path": cache_dir}},
            }
        )

        if len(parts) == 1:
            if not self.option("all"):
                raise RuntimeError(
                    "Add the --all option if you want to clear all "
                    "{} caches".format(parts[0])
                )

            if not os.path.exists(cache_dir):
                self.line("No cache entries for {}".format(parts[0]))
                return 0

            # Calculate number of entries
            entries_count = 0
            for path, dirs, files in os.walk(cache_dir):
                entries_count += len(files)

            delete = self.confirm(
                "<question>Delete {} entries?</>".format(entries_count)
            )
            if not delete:
                return 0

            cache.flush()
        elif len(parts) == 2:
            raise RuntimeError(
                "Only specifying the package name is not yet supported. "
                "Add a specific version to clear"
            )
        elif len(parts) == 3:
            package = parts[1]
            version = parts[2]

            if not cache.has("{}:{}".format(package, version)):
                self.line("No cache entries for {}:{}".format(package, version))
                return 0

            delete = self.confirm("Delete cache entry {}:{}".format(package, version))
            if not delete:
                return 0

            cache.forget("{}:{}".format(package, version))
        else:
            raise ValueError("Invalid cache key")




"""
A converter of python values to TOML Token instances.
"""
import datetime
import re

from poetry.utils._compat import basestring

from .. import tokens
from ..errors import TOMLError
from ..tokens import Token
from ..util import chunkate_string


class NotPrimitiveError(TOMLError):
    pass


_operator_tokens_by_type = {
    tokens.TYPE_OP_SQUARE_LEFT_BRACKET: tokens.Token(
        tokens.TYPE_OP_SQUARE_LEFT_BRACKET, u"["
    ),
    tokens.TYPE_OP_SQUARE_RIGHT_BRACKET: tokens.Token(
        tokens.TYPE_OP_SQUARE_RIGHT_BRACKET, u"]"
    ),
    tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET: tokens.Token(
        tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET, u"[["
    ),
    tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET: tokens.Token(
        tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET, u"]]"
    ),
    tokens.TYPE_OP_COMMA: tokens.Token(tokens.TYPE_OP_COMMA, u","),
    tokens.TYPE_NEWLINE: tokens.Token(tokens.TYPE_NEWLINE, u"\n"),
    tokens.TYPE_OPT_DOT: tokens.Token(tokens.TYPE_OPT_DOT, u"."),
}


def operator_token(token_type):
    return _operator_tokens_by_type[token_type]


def create_primitive_token(value, multiline_strings_allowed=True):
    """
    Creates and returns a single token for the given primitive atomic value.

    Raises NotPrimitiveError when the given value is not a primitive atomic value
    """
    if value is None:
        return create_primitive_token("")
    elif isinstance(value, bool):
        return tokens.Token(tokens.TYPE_BOOLEAN, u"true" if value else u"false")
    elif isinstance(value, int):
        return tokens.Token(tokens.TYPE_INTEGER, u"{}".format(value))
    elif isinstance(value, float):
        return tokens.Token(tokens.TYPE_FLOAT, u"{}".format(value))
    elif isinstance(value, (datetime.datetime, datetime.date, datetime.time)):
        return tokens.Token(tokens.TYPE_DATE, value.isoformat())
    elif isinstance(value, basestring):
        return create_string_token(
            value, multiline_strings_allowed=multiline_strings_allowed
        )

    raise NotPrimitiveError("{} of type {}".format(value, type(value)))


_bare_string_regex = re.compile("^[a-zA-Z0-9_-]*$")


def create_string_token(
    text, bare_string_allowed=False, multiline_strings_allowed=True
):
    """
    Creates and returns a single string token.

    Raises ValueError on non-string input.
    """
    if not isinstance(text, basestring):
        raise ValueError("Given value must be a string")

    if text == "":
        return tokens.Token(
            tokens.TYPE_STRING, '""'.format(_escape_single_line_quoted_string(text))
        )
    elif bare_string_allowed and _bare_string_regex.match(text):
        return tokens.Token(tokens.TYPE_BARE_STRING, text)
    elif multiline_strings_allowed and (
        len(tuple(c for c in text if c == "\n")) >= 2 or len(text) > 80
    ):
        # If containing two or more newlines or is longer than 80 characters we'll use the multiline string format
        return _create_multiline_string_token(text)
    else:
        return tokens.Token(
            tokens.TYPE_STRING, u'"{}"'.format(_escape_single_line_quoted_string(text))
        )


def _escape_single_line_quoted_string(text):
    return text.replace('"', '\\"')


def _create_multiline_string_token(text):
    escaped = text.replace(u'"""', u'"""')
    if len(escaped) > 50:
        return tokens.Token(
            tokens.TYPE_MULTILINE_STRING,
            u'"""\n{}\\\n"""'.format(_break_long_text(escaped)),
        )
    else:
        return tokens.Token(tokens.TYPE_MULTILINE_STRING, u'"""{}"""'.format(escaped))


def _break_long_text(text, maximum_length=75):
    """
    Breaks into lines of 75 character maximum length that are terminated by a backslash.
    """

    def next_line(remaining_text):

        # Returns a line and the remaining text

        if "\n" in remaining_text and remaining_text.index("\n") < maximum_length:
            i = remaining_text.index("\n")
            return remaining_text[: i + 1], remaining_text[i + 2 :]
        elif len(remaining_text) > maximum_length and " " in remaining_text:
            i = remaining_text[:maximum_length].rfind(" ")
            return remaining_text[: i + 1] + "\\\n", remaining_text[i + 2 :]
        else:
            return remaining_text, ""

    remaining_text = text
    lines = []
    while remaining_text:
        line, remaining_text = next_line(remaining_text)
        lines += [line]

    return "".join(lines)


def create_whitespace(source_substring):
    return Token(tokens.TYPE_WHITESPACE, source_substring)


def create_multiline_string(text, maximum_line_length=120):
    def escape(t):
        return t.replace(u'"""', u'"""')

    source_substring = u'"""\n{}"""'.format(
        u"\\\n".join(chunkate_string(escape(text), maximum_line_length))
    )
    return Token(tokens.TYPE_MULTILINE_STRING, source_substring)



from .builders import CompleteBuilder
from .builders import SdistBuilder
from .builders import WheelBuilder


class Builder:

    _FORMATS = {"sdist": SdistBuilder, "wheel": WheelBuilder, "all": CompleteBuilder}

    def __init__(self, poetry, venv, io):
        self._poetry = poetry
        self._venv = venv
        self._io = io

    def build(self, fmt):
        if fmt not in self._FORMATS:
            raise ValueError("Invalid format: {}".format(fmt))

        builder = self._FORMATS[fmt](self._poetry, self._venv, self._io)

        return builder.build()



import re

from typing import List

from ..command import Command


class DebugResolveCommand(Command):
    """
    Debugs dependency resolution.

    debug:resolve
        { package?* : packages to resolve. }
        { --E|extras=* : Extras to activate for the dependency. }
        { --python= : Python version(s) to use for resolution. }
    """

    _loggers = ["poetry.repositories.pypi_repository"]

    def handle(self):
        from poetry.packages import Dependency
        from poetry.packages import ProjectPackage
        from poetry.puzzle import Solver
        from poetry.repositories.repository import Repository
        from poetry.semver import parse_constraint

        packages = self.argument("package")

        if not packages:
            package = self.poetry.package
        else:
            requirements = self._determine_requirements(packages)
            requirements = self._format_requirements(requirements)

            # validate requirements format
            for constraint in requirements.values():
                parse_constraint(constraint)

            dependencies = []
            for name, constraint in requirements.items():
                dep = Dependency(name, constraint)
                extras = []
                for extra in self.option("extras"):
                    if " " in extra:
                        extras += [e.strip() for e in extra.split(" ")]
                    else:
                        extras.append(extra)

                for ex in extras:
                    dep.extras.append(ex)

                dependencies.append(dep)

            package = ProjectPackage(
                self.poetry.package.name, self.poetry.package.version
            )

            package.python_versions = (
                self.option("python") or self.poetry.package.python_versions
            )
            for dep in dependencies:
                package.requires.append(dep)

        solver = Solver(
            package, self.poetry.pool, Repository(), Repository(), self.output
        )

        ops = solver.solve()

        self.line("")
        self.line("Resolution results:")
        self.line("")

        for op in ops:
            package = op.package
            self.line(
                "  - <info>{}</info> (<comment>{}</comment>)".format(
                    package.name, package.version
                )
            )
            if package.requirements:
                for req_name, req_value in package.requirements.items():
                    self.line("    - {}: {}".format(req_name, req_value))

    def _determine_requirements(self, requires):  # type: (List[str]) -> List[str]
        if not requires:
            return []

        requires = self._parse_name_version_pairs(requires)
        result = []
        for requirement in requires:
            if "version" not in requirement:
                requirement["version"] = "*"

            result.append("{} {}".format(requirement["name"], requirement["version"]))

        return result

    def _parse_name_version_pairs(self, pairs):  # type: (list) -> list
        result = []

        for i in range(len(pairs)):
            pair = re.sub("^([^=: ]+)[=: ](.*)$", "\\1 \\2", pairs[i].strip())
            pair = pair.strip()

            if " " in pair:
                name, version = pair.split(" ", 2)
                result.append({"name": name, "version": version})
            else:
                result.append({"name": pair})

        return result

    def _format_requirements(self, requirements):  # type: (List[str]) -> dict
        requires = {}
        requirements = self._parse_name_version_pairs(requirements)
        for requirement in requirements:
            requires[requirement["name"]] = requirement["version"]

        return requires



from typing import List

from poetry.utils._compat import Path


class Include(object):
    """
    Represents an "include" entry.

    It can be a glob string, a single file or a directory.

    This class will then detect the type of this include:

        - a package
        - a module
        - a file
        - a directory
    """

    def __init__(self, base, include):  # type: (Path, str) -> None
        self._base = base
        self._include = include

        self._elements = sorted(list(self._base.glob(self._include)))

    @property
    def base(self):  # type: () -> Path
        return self._base

    @property
    def elements(self):  # type: () -> List[Path]
        return self._elements

    def is_empty(self):  # type: () -> bool
        return len(self._elements) == 0

    def refresh(self):  # type: () -> Include
        self._elements = sorted(list(self._base.glob(self._include)))

        return self




"""
A regular expression based Lexer/tokenizer for TOML.
"""

from collections import namedtuple
import re
from .. import tokens
from ..errors import TOMLError

TokenSpec = namedtuple("TokenSpec", ("type", "re"))

# Specs of all the valid tokens
_LEXICAL_SPECS = (
    TokenSpec(tokens.TYPE_COMMENT, re.compile(r"^(#.*)\n")),
    TokenSpec(
        tokens.TYPE_STRING, re.compile(r'^("(([^"]|\\")+?[^\\]|([^"]|\\")|)")')
    ),  # Single line only
    TokenSpec(tokens.TYPE_MULTILINE_STRING, re.compile(r'^(""".*?""")', re.DOTALL)),
    TokenSpec(tokens.TYPE_LITERAL_STRING, re.compile(r"^('.*?')")),
    TokenSpec(
        tokens.TYPE_MULTILINE_LITERAL_STRING, re.compile(r"^('''.*?''')", re.DOTALL)
    ),
    TokenSpec(tokens.TYPE_BARE_STRING, re.compile(r"^([A-Za-z0-9_-]+)")),
    TokenSpec(
        tokens.TYPE_DATE,
        re.compile(
            r"^([0-9]{4}-[0-9]{2}-[0-9]{2}(T[0-9]{2}:[0-9]{2}:[0-9]{2}(\.[0-9]*)?)?(([zZ])|((\+|-)[0-9]{2}:[0-9]{2}))?)"
        ),
    ),
    TokenSpec(tokens.TYPE_WHITESPACE, re.compile(r"^( |\t)", re.DOTALL)),
    TokenSpec(tokens.TYPE_INTEGER, re.compile(r"^(((\+|-)[0-9_]+)|([0-9][0-9_]*))")),
    TokenSpec(
        tokens.TYPE_FLOAT,
        re.compile(
            r"^((((\+|-)[0-9_]+)|([1-9][0-9_]*))(\.[0-9_]+)?([eE](\+|-)?[0-9_]+)?)"
        ),
    ),
    TokenSpec(tokens.TYPE_BOOLEAN, re.compile(r"^(true|false)")),
    TokenSpec(tokens.TYPE_OP_SQUARE_LEFT_BRACKET, re.compile(r"^(\[)")),
    TokenSpec(tokens.TYPE_OP_SQUARE_RIGHT_BRACKET, re.compile(r"^(\])")),
    TokenSpec(tokens.TYPE_OP_CURLY_LEFT_BRACKET, re.compile(r"^(\{)")),
    TokenSpec(tokens.TYPE_OP_CURLY_RIGHT_BRACKET, re.compile(r"^(\})")),
    TokenSpec(tokens.TYPE_OP_ASSIGNMENT, re.compile(r"^(=)")),
    TokenSpec(tokens.TYPE_OP_COMMA, re.compile(r"^(,)")),
    TokenSpec(tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET, re.compile(r"^(\[\[)")),
    TokenSpec(tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET, re.compile(r"^(\]\])")),
    TokenSpec(tokens.TYPE_OPT_DOT, re.compile(r"^(\.)")),
    TokenSpec(tokens.TYPE_NEWLINE, re.compile("^(\n|\r\n)")),
)


def _next_token_candidates(source):
    matches = []
    for token_spec in _LEXICAL_SPECS:
        match = token_spec.re.search(source)
        if match:
            matches.append(tokens.Token(token_spec.type, match.group(1)))
    return matches


def _choose_from_next_token_candidates(candidates):

    if len(candidates) == 1:
        return candidates[0]
    elif len(candidates) > 1:
        # Return the maximal-munch with ties broken by natural order of token type.
        maximal_munch_length = max(len(token.source_substring) for token in candidates)
        maximal_munches = [
            token
            for token in candidates
            if len(token.source_substring) == maximal_munch_length
        ]
        return sorted(maximal_munches)[0]  # Return the first in sorting by priority


def _munch_a_token(source):
    """
    Munches a single Token instance if it could recognize one at the beginning of the
    given source text, or None if no token type could be recognized.
    """
    candidates = _next_token_candidates(source)
    return _choose_from_next_token_candidates(candidates)


class LexerError(TOMLError):
    def __init__(self, message):
        self._message = message

    def __repr__(self):
        return self._message

    def __str__(self):
        return self._message


def tokenize(source, is_top_level=False):
    """
    Tokenizes the input TOML source into a stream of tokens.

    If is_top_level is set to True, will make sure that the input source has a trailing newline character
    before it is tokenized.

    Raises a LexerError when it fails recognize another token while not at the end of the source.
    """

    # Newlines are going to be normalized to UNIX newlines.
    source = source.replace("\r\n", "\n")

    if is_top_level and source and source[-1] != "\n":
        source += "\n"

    next_row = 1
    next_col = 1
    next_index = 0

    while next_index < len(source):

        new_token = _munch_a_token(source[next_index:])

        if not new_token:
            raise LexerError(
                "failed to read the next token at ({}, {}): {}".format(
                    next_row, next_col, source[next_index:]
                )
            )

        # Set the col and row on the new token
        new_token = tokens.Token(
            new_token.type, new_token.source_substring, next_col, next_row
        )

        # Advance the index, row and col count
        next_index += len(new_token.source_substring)
        for c in new_token.source_substring:
            if c == "\n":
                next_row += 1
                next_col = 1
            else:
                next_col += 1

        yield new_token



from .venv_command import VenvCommand


class RunCommand(VenvCommand):
    """
    Runs a command in the appropriate environment.

    run
        { args* : The command and arguments/options to run. }
    """

    def handle(self):
        args = self.argument("args")
        script = args[0]
        scripts = self.poetry.local_config.get("scripts")

        if scripts and script in scripts:
            return self.run_script(scripts[script], args)

        venv = self.venv

        return venv.execute(*args)

    def run_script(self, script, args):
        module, callable_ = script.split(":")

        src_in_sys_path = "sys.path.append('src'); " if self._module.is_in_src() else ""

        cmd = ["python", "-c"]

        cmd += [
            '"import sys; '
            "from importlib import import_module; "
            "sys.argv = {!r}; {}"
            "import_module('{}').{}()\"".format(
                args, src_in_sys_path, module, callable_
            )
        ]

        return self.venv.run(*cmd, shell=True, call=True)

    @property
    def _module(self):
        from ...masonry.utils.module import Module

        poetry = self.poetry
        package = poetry.package
        path = poetry.file.parent
        module = Module(package.name, path.as_posix())
        return module

    def merge_application_definition(self, merge_args=True):
        if self._application is None or (
            self._application_definition_merged
            and (self._application_definition_merged_with_args or not merge_args)
        ):
            return

        if merge_args:
            current_arguments = self._definition.get_arguments()
            self._definition.set_arguments(
                self._application.get_definition().get_arguments()
            )
            self._definition.add_arguments(current_arguments)

        self._application_definition_merged = True
        if merge_args:
            self._application_definition_merged_with_args = True



class InvalidElementError(Exception):
    """
    Raised by Element factories when the given sequence of tokens or sub-elements are invalid for the
    specific type of Element being created.
    """

    def __init__(self, message):
        self.message = message

    def __repr__(self):
        return "InvalidElementError: {}".format(self.message)



class BaseConstraint(object):
    def matches(self, provider):
        raise NotImplementedError()

    def allows_all(self, other):
        raise NotImplementedError()

    def allows_any(self, other):
        raise NotImplementedError()

    def difference(self, other):
        raise NotImplementedError()

    def intersect(self, other):
        raise NotImplementedError()

    def is_empty(self):
        return False



from ..errors import TOMLError


class ParsingError(TOMLError):
    def __init__(self, message="", token=None):
        self.message = message
        self.token = token

    def __repr__(self):
        if self.message and self.token:
            return "{} at row {} and col {}".format(
                self.message, self.token.row, self.token.col
            )
        else:
            return self.message

    def __str__(self):
        return repr(self)



"""
Generate and work with PEP 425 Compatibility Tags.

Base implementation taken from
https://github.com/pypa/wheel/blob/master/wheel/pep425tags.py
and adapted to work with poetry's venv util.
"""
from __future__ import unicode_literals

import distutils.util
import sys
import warnings


def get_abbr_impl(venv):
    """Return abbreviated implementation name."""
    impl = venv.python_implementation

    if impl == "PyPy":
        return "pp"
    elif impl == "Jython":
        return "jy"
    elif impl == "IronPython":
        return "ip"
    elif impl == "CPython":
        return "cp"

    raise LookupError("Unknown Python implementation: " + impl)


def get_impl_ver(venv):
    """Return implementation version."""
    impl_ver = venv.config_var("py_version_nodot")
    if not impl_ver or get_abbr_impl(venv) == "pp":
        impl_ver = "".join(map(str, get_impl_version_info(venv)))

    return impl_ver


def get_impl_version_info(venv):
    """Return sys.version_info-like tuple for use in decrementing the minor
    version."""
    if get_abbr_impl(venv) == "pp":
        # as per https://github.com/pypa/pip/issues/2882
        return venv.version_info[:3]
    else:
        return venv.version_info[:2]


def get_flag(venv, var, fallback, expected=True, warn=True):
    """Use a fallback method for determining SOABI flags if the needed config
    var is unset or unavailable."""
    val = venv.config_var(var)
    if val is None:
        if warn:
            warnings.warn(
                "Config variable '{0}' is unset, Python ABI tag may "
                "be incorrect".format(var),
                RuntimeWarning,
                2,
            )
        return fallback()
    return val == expected


def get_abi_tag(venv):
    """Return the ABI tag based on SOABI (if available) or emulate SOABI
    (CPython 2, PyPy)."""
    soabi = venv.config_var("SOABI")
    impl = get_abbr_impl(venv)
    if not soabi and impl in ("cp", "pp") and hasattr(sys, "maxunicode"):
        d = ""
        m = ""
        u = ""
        if get_flag(
            venv,
            "Py_DEBUG",
            lambda: hasattr(sys, "gettotalrefcount"),
            warn=(impl == "cp"),
        ):
            d = "d"
        if get_flag(venv, "WITH_PYMALLOC", lambda: impl == "cp", warn=(impl == "cp")):
            m = "m"
        if get_flag(
            venv,
            "Py_UNICODE_SIZE",
            lambda: sys.maxunicode == 0x10ffff,
            expected=4,
            warn=(impl == "cp" and venv.version_info < (3, 3)),
        ) and venv.version_info < (3, 3):
            u = "u"
        abi = "%s%s%s%s%s" % (impl, get_impl_ver(venv), d, m, u)
    elif soabi and soabi.startswith("cpython-"):
        abi = "cp" + soabi.split("-")[1]
    elif soabi:
        abi = soabi.replace(".", "_").replace("-", "_")
    else:
        abi = None
    return abi


def get_platform():
    """Return our platform name 'win32', 'linux_x86_64'"""
    # XXX remove distutils dependency
    result = distutils.util.get_platform().replace(".", "_").replace("-", "_")
    if result == "linux_x86_64" and sys.maxsize == 2147483647:
        # pip pull request #3497
        result = "linux_i686"
    return result


def get_supported(venv, versions=None, supplied_platform=None):
    """Return a list of supported tags for each version specified in
    `versions`.
    :param versions: a list of string versions, of the form ["33", "32"],
        or None. The first version will be assumed to support our ABI.
    """
    supported = []

    # Versions must be given with respect to the preference
    if versions is None:
        versions = []
        version_info = get_impl_version_info(venv)
        major = version_info[:-1]
        # Support all previous minor Python versions.
        for minor in range(version_info[-1], -1, -1):
            versions.append("".join(map(str, major + (minor,))))

    impl = get_abbr_impl(venv)

    abis = []

    abi = get_abi_tag(venv)
    if abi:
        abis[0:0] = [abi]

    abi3s = set()
    import imp

    for suffix in imp.get_suffixes():
        if suffix[0].startswith(".abi"):
            abi3s.add(suffix[0].split(".", 2)[1])

    abis.extend(sorted(list(abi3s)))

    abis.append("none")

    platforms = []
    if supplied_platform:
        platforms.append(supplied_platform)
    platforms.append(get_platform())

    # Current version, current API (built specifically for our Python):
    for abi in abis:
        for arch in platforms:
            supported.append(("%s%s" % (impl, versions[0]), abi, arch))

    # abi3 modules compatible with older version of Python
    for version in versions[1:]:
        # abi3 was introduced in Python 3.2
        if version in ("31", "30"):
            break
        for abi in abi3s:  # empty set if not Python 3
            for arch in platforms:
                supported.append(("%s%s" % (impl, version), abi, arch))

    # No abi / arch, but requires our implementation:
    for i, version in enumerate(versions):
        supported.append(("%s%s" % (impl, version), "none", "any"))
        if i == 0:
            # Tagged specifically as being cross-version compatible
            # (with just the major version specified)
            supported.append(("%s%s" % (impl, versions[0][0]), "none", "any"))

    # Major Python version + platform; e.g. binaries not using the Python API
    supported.append(("py%s" % (versions[0][0]), "none", arch))

    # No abi / arch, generic Python
    for i, version in enumerate(versions):
        supported.append(("py%s" % (version,), "none", "any"))
        if i == 0:
            supported.append(("py%s" % (version[0]), "none", "any"))

    return supported




"""
    A Recursive Descent implementation of a lexical parser for TOML.

    Grammar:
    --------

    Newline -> NEWLINE
    Comment -> COMMENT Newline
    LineTerminator -> Comment | Newline
    Space -> WHITESPACE Space | WHITESPACE | EMPTY
    TableHeader -> Space [ Space TableHeaderName Space ] Space LineTerminator |
        Space [[ Space TableHeaderName Space ]] Space LineTerminator
    TableHeaderName -> STRING Space '.' Space TableHeaderName | STRING
    Atomic -> STRING | INTEGER | FLOAT | DATE | BOOLEAN

    Array -> '[' Space ArrayInternal Space ']' | '[' Space ArrayInternal Space LineTerminator Space ']'
    ArrayInternal -> LineTerminator Space ArrayInternal | Value Space ',' Space LineTerminator Space ArrayInternal |
        Value Space ',' Space ArrayInternal | LineTerminator | Value | EMPTY

    InlineTable -> '{' Space InlineTableInternal Space '}'
    InlineTableKeyValuePair = STRING Space '=' Space Value
    InlineTableInternal -> InlineTableKeyValuePair Space ',' Space InlineTableInternal |
        InlineTableKeyValuePair | Empty

    Value -> Atomic | InlineTable | Array
    KeyValuePair -> Space STRING Space '=' Space Value Space LineTerminator

    TableBody -> KeyValuePair TableBody | EmptyLine TableBody | EmptyLine | KeyValuePair

    EmptyLine -> Space LineTerminator
    FileEntry -> TableHeader | TableBody

    TOMLFileElements -> FileEntry TOMLFileElements | FileEntry | EmptyLine | EMPTY
"""

from ..elements.array import ArrayElement
from ..elements.atomic import AtomicElement
from ..elements.inlinetable import InlineTableElement
from ..elements.metadata import (
    NewlineElement,
    CommentElement,
    WhitespaceElement,
    PunctuationElement,
)
from ..elements.table import TableElement
from ..elements.tableheader import TableHeaderElement
from ..tokens import TYPE_BARE_STRING
from ..tokens import TYPE_BOOLEAN
from ..tokens import TYPE_COMMENT
from ..tokens import TYPE_DATE
from ..tokens import TYPE_FLOAT
from ..tokens import TYPE_INTEGER
from ..tokens import TYPE_LITERAL_STRING
from ..tokens import TYPE_MULTILINE_LITERAL_STRING
from ..tokens import TYPE_MULTILINE_STRING
from ..tokens import TYPE_NEWLINE
from ..tokens import TYPE_OP_ASSIGNMENT
from ..tokens import TYPE_OP_COMMA
from ..tokens import TYPE_OP_CURLY_LEFT_BRACKET
from ..tokens import TYPE_OP_CURLY_RIGHT_BRACKET
from ..tokens import TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET
from ..tokens import TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET
from ..tokens import TYPE_OP_SQUARE_LEFT_BRACKET
from ..tokens import TYPE_OP_SQUARE_RIGHT_BRACKET
from ..tokens import TYPE_OPT_DOT
from ..tokens import TYPE_STRING
from ..tokens import TYPE_WHITESPACE

from .recdesc import capture_from
from .errors import ParsingError

"""
    Non-terminals are represented as functions which return (RESULT, pending_token_stream), or raise ParsingError.
"""


def token(token_type):
    def factory(ts):
        t = ts.head
        if t.type != token_type:
            raise ParsingError("Expected a token of type {}".format(token_type))
        return t, ts.tail

    return factory


def newline_element(token_stream):
    """
    Returns NewlineElement, pending_token_stream or raises ParsingError.
    """
    captured = capture_from(token_stream).find(token(TYPE_NEWLINE))
    return NewlineElement(captured.value()), captured.pending_tokens


def comment_tokens(ts1):
    c1 = capture_from(ts1).find(token(TYPE_COMMENT)).and_find(token(TYPE_NEWLINE))
    return c1.value(), c1.pending_tokens


def comment_element(token_stream):
    """
    Returns CommentElement, pending_token_stream or raises ParsingError.
    """
    captured = capture_from(token_stream).find(comment_tokens)
    return CommentElement(captured.value()), captured.pending_tokens


def line_terminator_tokens(token_stream):
    captured = (
        capture_from(token_stream).find(comment_tokens).or_find(token(TYPE_NEWLINE))
    )
    return captured.value(), captured.pending_tokens


def line_terminator_element(token_stream):
    captured = capture_from(token_stream).find(comment_element).or_find(newline_element)
    return captured.value("Expected a comment or a newline")[0], captured.pending_tokens


def zero_or_more_tokens(token_type):
    def factory(token_stream):
        def more(ts):
            c = (
                capture_from(ts)
                .find(token(token_type))
                .and_find(zero_or_more_tokens(token_type))
            )
            return c.value(), c.pending_tokens

        def two(ts):
            c = capture_from(ts).find(token(TYPE_WHITESPACE))
            return c.value(), c.pending

        def zero(ts):
            return tuple(), ts

        captured = capture_from(token_stream).find(more).or_find(two).or_find(zero)
        return captured.value(), captured.pending_tokens

    return factory


def space_element(token_stream):
    captured = capture_from(token_stream).find(zero_or_more_tokens(TYPE_WHITESPACE))
    return (
        WhitespaceElement([t for t in captured.value() if t]),
        captured.pending_tokens,
    )


def string_token(token_stream):
    captured = (
        capture_from(token_stream)
        .find(token(TYPE_BARE_STRING))
        .or_find(token(TYPE_STRING))
        .or_find(token(TYPE_LITERAL_STRING))
        .or_find(token(TYPE_MULTILINE_STRING))
        .or_find(token(TYPE_MULTILINE_LITERAL_STRING))
    )
    return captured.value("Expected a string"), captured.pending_tokens


def string_element(token_stream):
    captured = capture_from(token_stream).find(string_token)
    return AtomicElement(captured.value()), captured.pending_tokens


def table_header_name_tokens(token_stream):
    def one(ts):
        c = (
            capture_from(ts)
            .find(string_token)
            .and_find(zero_or_more_tokens(TYPE_WHITESPACE))
            .and_find(token(TYPE_OPT_DOT))
            .and_find(zero_or_more_tokens(TYPE_WHITESPACE))
            .and_find(table_header_name_tokens)
        )
        return c.value(), c.pending_tokens

    captured = capture_from(token_stream).find(one).or_find(string_token)
    return captured.value(), captured.pending_tokens


def table_header_element(token_stream):
    def single(ts1):
        c1 = (
            capture_from(ts1)
            .find(zero_or_more_tokens(TYPE_WHITESPACE))
            .and_find(token(TYPE_OP_SQUARE_LEFT_BRACKET))
            .and_find(zero_or_more_tokens(TYPE_WHITESPACE))
            .and_find(table_header_name_tokens)
            .and_find(zero_or_more_tokens(TYPE_WHITESPACE))
            .and_find(token(TYPE_OP_SQUARE_RIGHT_BRACKET))
            .and_find(zero_or_more_tokens(TYPE_WHITESPACE))
            .and_find(line_terminator_tokens)
        )

        return c1.value(), c1.pending_tokens

    def double(ts2):
        c2 = (
            capture_from(ts2)
            .find(zero_or_more_tokens(TYPE_WHITESPACE))
            .and_find(token(TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET))
            .and_find(zero_or_more_tokens(TYPE_WHITESPACE))
            .and_find(table_header_name_tokens)
            .and_find(zero_or_more_tokens(TYPE_WHITESPACE))
            .and_find(token(TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET))
            .and_find(zero_or_more_tokens(TYPE_WHITESPACE))
            .and_find(line_terminator_tokens)
        )

        return c2.value(), c2.pending_tokens

    captured = capture_from(token_stream).find(single).or_find(double)
    return TableHeaderElement(captured.value()), captured.pending_tokens


def atomic_element(token_stream):
    captured = (
        capture_from(token_stream)
        .find(string_token)
        .or_find(token(TYPE_INTEGER))
        .or_find(token(TYPE_FLOAT))
        .or_find(token(TYPE_DATE))
        .or_find(token(TYPE_BOOLEAN))
    )
    return (
        AtomicElement(captured.value("Expected an atomic primitive value")),
        captured.pending_tokens,
    )


def punctuation_element(token_type):
    def factory(ts):
        c = capture_from(ts).find(token(token_type))
        return (
            PunctuationElement(
                c.value("Expected the punctuation element: {}".format(token_type))
            ),
            c.pending_tokens,
        )

    return factory


def value(token_stream):
    captured = (
        capture_from(token_stream)
        .find(atomic_element)
        .or_find(array_element)
        .or_find(inline_table_element)
    )
    return (
        captured.value("Expected a primitive value, array or an inline table"),
        captured.pending_tokens,
    )


def array_internal(ts):
    def zero(ts0):
        c = (
            capture_from(ts0)
            .and_find(line_terminator_element)
            .and_find(space_element)
            .and_find(array_internal)
        )
        return c.value(), c.pending_tokens

    def one(ts1):
        c = (
            capture_from(ts1)
            .find(value)
            .and_find(space_element)
            .and_find(punctuation_element(TYPE_OP_COMMA))
            .and_find(space_element)
            .and_find(line_terminator_element)
            .and_find(space_element)
            .and_find(array_internal)
        )
        return c.value(), c.pending_tokens

    def two(ts2):
        c = (
            capture_from(ts2)
            .find(value)
            .and_find(space_element)
            .and_find(punctuation_element(TYPE_OP_COMMA))
            .and_find(space_element)
            .and_find(array_internal)
        )
        return c.value(), c.pending_tokens

    def three(ts3):
        c = capture_from(ts3).find(space_element).and_find(line_terminator_element)
        return c.value(), c.pending_tokens

    captured = (
        capture_from(ts)
        .find(zero)
        .or_find(one)
        .or_find(two)
        .or_find(three)
        .or_find(value)
        .or_empty()
    )
    return captured.value(), captured.pending_tokens


def array_element(token_stream):
    def one(ts1):
        ca = (
            capture_from(ts1)
            .find(punctuation_element(TYPE_OP_SQUARE_LEFT_BRACKET))
            .and_find(space_element)
            .and_find(array_internal)
            .and_find(space_element)
            .and_find(punctuation_element(TYPE_OP_SQUARE_RIGHT_BRACKET))
        )
        return ca.value(), ca.pending_tokens

    def two(ts2):
        ca = (
            capture_from(ts2)
            .find(punctuation_element(TYPE_OP_SQUARE_LEFT_BRACKET))
            .and_find(space_element)
            .and_find(array_internal)
            .and_find(space_element)
            .and_find(line_terminator_element)
            .and_find(space_element)
            .and_find(punctuation_element(TYPE_OP_SQUARE_RIGHT_BRACKET))
        )
        return ca.value(), ca.pending_tokens

    captured = capture_from(token_stream).find(one).or_find(two)
    return ArrayElement(captured.value()), captured.pending_tokens


def inline_table_element(token_stream):
    # InlineTableElement -> '{' Space InlineTableInternal Space '}'
    # InlineTableKeyValuePair = STRING Space '=' Space Value
    # InlineTableInternal -> InlineTableKeyValuePair Space ',' Space InlineTableInternal |
    #     InlineTableKeyValuePair | Empty
    def key_value(ts):
        ca = (
            capture_from(ts)
            .find(string_element)
            .and_find(space_element)
            .and_find(punctuation_element(TYPE_OP_ASSIGNMENT))
            .and_find(space_element)
            .and_find(value)
        )
        return ca.value(), ca.pending_tokens

    def internal(ts):
        def one(ts1):
            c1 = (
                capture_from(ts1)
                .find(key_value)
                .and_find(space_element)
                .and_find(punctuation_element(TYPE_OP_COMMA))
                .and_find(space_element)
                .and_find(internal)
            )
            return c1.value(), c1.pending_tokens

        c = capture_from(ts).find(one).or_find(key_value).or_empty()
        return c.value(), c.pending_tokens

    captured = (
        capture_from(token_stream)
        .find(punctuation_element(TYPE_OP_CURLY_LEFT_BRACKET))
        .and_find(space_element)
        .and_find(internal)
        .and_find(space_element)
        .and_find(punctuation_element(TYPE_OP_CURLY_RIGHT_BRACKET))
    )

    return InlineTableElement(captured.value()), captured.pending_tokens


def key_value_pair(token_stream):
    captured = (
        capture_from(token_stream)
        .find(space_element)
        .and_find(string_element)
        .and_find(space_element)
        .and_find(punctuation_element(TYPE_OP_ASSIGNMENT))
        .and_find(space_element)
        .and_find(value)
        .and_find(space_element)
        .and_find(line_terminator_element)
    )
    return captured.value(), captured.pending_tokens


def table_body_elements(token_stream):
    # TableBody -> KeyValuePair TableBody | EmptyLine TableBody | EmptyLine | KeyValuePair
    def one(ts1):
        c = capture_from(ts1).find(key_value_pair).and_find(table_body_elements)
        return c.value(), c.pending_tokens

    def two(ts2):
        c = capture_from(ts2).find(empty_line_elements).and_find(table_body_elements)
        return c.value(), c.pending_tokens

    captured = (
        capture_from(token_stream)
        .find(one)
        .or_find(two)
        .or_find(empty_line_elements)
        .or_find(key_value_pair)
    )

    return captured.value(), captured.pending_tokens


def table_body_element(token_stream):
    captured = capture_from(token_stream).find(table_body_elements)
    return TableElement(captured.value()), captured.pending_tokens


def empty_line_tokens(ts1):
    c1 = capture_from(ts1).find(space_element).and_find(line_terminator_element)
    return c1.value(), c1.pending_tokens


def empty_line_elements(token_stream):
    captured = capture_from(token_stream).find(empty_line_tokens)
    return captured.value(), captured.pending_tokens


def file_entry_element(token_stream):
    captured = (
        capture_from(token_stream)
        .find(table_header_element)
        .or_find(table_body_element)
    )
    return captured.value(), captured.pending_tokens


def toml_file_elements(token_stream):
    def one(ts1):
        c1 = capture_from(ts1).find(file_entry_element).and_find(toml_file_elements)
        return c1.value(), c1.pending_tokens

    captured = (
        capture_from(token_stream).find(one).or_find(file_entry_element).or_empty()
    )
    return captured.value(), captured.pending_tokens



import re

from typing import List
from typing import Union

from .empty_constraint import EmptyConstraint
from .patterns import COMPLETE_VERSION
from .version_constraint import VersionConstraint
from .version_range import VersionRange
from .version_union import VersionUnion


class Version(VersionRange):
    """
    A parsed semantic version number.
    """

    def __init__(
        self,
        major,  # type: int
        minor=None,  # type: Union[int, None]
        patch=None,  # type: Union[int, None]
        pre=None,  # type: Union[str, None]
        build=None,  # type: Union[str, None]
        text=None,  # type: Union[str, None]
        precision=None,  # type: Union[str, None]
    ):  # type: () -> None
        self._major = int(major)
        self._precision = None
        if precision is None:
            self._precision = 1

        if minor is None:
            minor = 0
        else:
            if self._precision is not None:
                self._precision += 1

        self._minor = int(minor)

        if patch is None:
            patch = 0
        else:
            if self._precision is not None:
                self._precision += 1

        if precision is not None:
            self._precision = precision

        self._patch = int(patch)

        if text is None:
            parts = [str(major)]
            if self._precision >= 2 or minor != 0:
                parts.append(str(minor))

                if self._precision >= 3 or patch != 0:
                    parts.append(str(patch))

            text = ".".join(parts)
            if pre:
                text += "-{}".format(pre)

            if build:
                text += "+{}".format(build)

        self._text = text

        pre = self._normalize_prerelease(pre)

        self._prerelease = []
        if pre is not None:
            self._prerelease = self._split_parts(pre)

        build = self._normalize_build(build)

        self._build = []
        if build is not None:
            if build.startswith(("-", "+")):
                build = build[1:]

            self._build = self._split_parts(build)

    @property
    def major(self):  # type: () -> int
        return self._major

    @property
    def minor(self):  # type: () -> int
        return self._minor

    @property
    def patch(self):  # type: () -> int
        return self._patch

    @property
    def prerelease(self):  # type: () -> List[str]
        return self._prerelease

    @property
    def build(self):  # type: () -> List[str]
        return self._build

    @property
    def text(self):
        return self._text

    @property
    def stable(self):
        if not self.is_prerelease():
            return self

        return self.next_patch

    @property
    def next_major(self):  # type: () -> Version
        if self.is_prerelease() and self.minor == 0 and self.patch == 0:
            return Version(self.major, self.minor, self.patch)

        return self._increment_major()

    @property
    def next_minor(self):  # type: () -> Version
        if self.is_prerelease() and self.patch == 0:
            return Version(self.major, self.minor, self.patch)

        return self._increment_minor()

    @property
    def next_patch(self):  # type: () -> Version
        if self.is_prerelease():
            return Version(self.major, self.minor, self.patch)

        return self._increment_patch()

    @property
    def next_breaking(self):  # type: () -> Version
        if self.major == 0:
            if self.minor != 0:
                return self._increment_minor()

            if self._precision == 1:
                return self._increment_major()
            elif self._precision == 2:
                return self._increment_minor()

            return self._increment_patch()

        return self._increment_major()

    @property
    def first_prerelease(self):  # type: () -> Version
        return Version.parse(
            "{}.{}.{}-alpha.0".format(self.major, self.minor, self.patch)
        )

    @property
    def min(self):
        return self

    @property
    def max(self):
        return self

    @property
    def include_min(self):
        return True

    @property
    def include_max(self):
        return True

    @classmethod
    def parse(cls, text):  # type: (str) -> Version
        match = COMPLETE_VERSION.match(text)
        if match is None:
            raise ValueError('Unable to parse "{}".'.format(text))

        text = text.rstrip(".")

        major = int(match.group(1))
        minor = int(match.group(2)) if match.group(2) else None
        patch = int(match.group(3)) if match.group(3) else None

        pre = match.group(4)
        build = match.group(5)

        if build:
            build = build.lstrip("+")

        return Version(major, minor, patch, pre, build, text)

    def is_any(self):
        return False

    def is_empty(self):
        return False

    def is_prerelease(self):  # type: () -> bool
        return len(self._prerelease) > 0

    def allows(self, version):  # type: (Version) -> bool
        return self == version

    def allows_all(self, other):  # type: (VersionConstraint) -> bool
        return other.is_empty() or other == self

    def allows_any(self, other):  # type: (VersionConstraint) -> bool
        return other.allows(self)

    def intersect(self, other):  # type: (VersionConstraint) -> VersionConstraint
        if other.allows(self):
            return self

        return EmptyConstraint()

    def union(self, other):  # type: (VersionConstraint) -> VersionConstraint
        from .version_range import VersionRange

        if other.allows(self):
            return other

        if isinstance(other, VersionRange):
            if other.min == self:
                return VersionRange(
                    other.min,
                    other.max,
                    include_min=True,
                    include_max=other.include_max,
                )

            if other.max == self:
                return VersionRange(
                    other.min,
                    other.max,
                    include_min=other.include_min,
                    include_max=True,
                )

        return VersionUnion.of(self, other)

    def difference(self, other):  # type: (VersionConstraint) -> VersionConstraint
        if other.allows(self):
            return EmptyConstraint()

        return self

    def _increment_major(self):  # type: () -> Version
        return Version(self.major + 1, 0, 0, precision=self._precision)

    def _increment_minor(self):  # type: () -> Version
        return Version(self.major, self.minor + 1, 0, precision=self._precision)

    def _increment_patch(self):  # type: () -> Version
        return Version(
            self.major, self.minor, self.patch + 1, precision=self._precision
        )

    def _normalize_prerelease(self, pre):  # type: (str) -> str
        if not pre:
            return

        m = re.match("(?i)^(a|alpha|b|beta|c|pre|rc|dev)[-.]?(\d+)?$", pre)
        if not m:
            return

        modifier = m.group(1)
        number = m.group(2)

        if number is None:
            number = 0

        if modifier == "a":
            modifier = "alpha"
        elif modifier == "b":
            modifier = "beta"
        elif modifier in {"c", "pre"}:
            modifier = "rc"
        elif modifier == "dev":
            modifier = "alpha"

        return "{}.{}".format(modifier, number)

    def _normalize_build(self, build):  # type: (str) -> str
        if not build:
            return

        if build == "0":
            return

        if build.startswith("post"):
            build = build.lstrip("post")

        if not build:
            return

        return build

    def _split_parts(self, text):  # type: (str) -> List[Union[str, int]]
        parts = text.split(".")

        for i, part in enumerate(parts):
            try:
                parts[i] = int(part)
            except (TypeError, ValueError):
                continue

        return parts

    def __lt__(self, other):
        return self._cmp(other) < 0

    def __le__(self, other):
        return self._cmp(other) <= 0

    def __gt__(self, other):
        return self._cmp(other) > 0

    def __ge__(self, other):
        return self._cmp(other) >= 0

    def _cmp(self, other):
        if not isinstance(other, VersionConstraint):
            return NotImplemented

        if not isinstance(other, Version):
            return -other._cmp(self)

        if self.major != other.major:
            return self._cmp_parts(self.major, other.major)

        if self.minor != other.minor:
            return self._cmp_parts(self.minor, other.minor)

        if self.patch != other.patch:
            return self._cmp_parts(self.patch, other.patch)

        # Pre-releases always come before no pre-release string.
        if not self.is_prerelease() and other.is_prerelease():
            return 1

        if not other.is_prerelease() and self.is_prerelease():
            return -1

        comparison = self._cmp_lists(self.prerelease, other.prerelease)
        if comparison != 0:
            return comparison

        # Builds always come after no build string.
        if not self.build and other.build:
            return -1

        if not other.build and self.build:
            return 1

        return self._cmp_lists(self.build, other.build)

    def _cmp_parts(self, a, b):
        if a < b:
            return -1
        elif a > b:
            return 1

        return 0

    def _cmp_lists(self, a, b):  # type: (List, List) -> int
        for i in range(max(len(a), len(b))):
            a_part = None
            if i < len(a):
                a_part = a[i]

            b_part = None
            if i < len(b):
                b_part = b[i]

            if a_part == b_part:
                continue

            # Missing parts come before present ones.
            if a_part is None:
                return -1

            if b_part is None:
                return 1

            if isinstance(a_part, int):
                if isinstance(b_part, int):
                    return self._cmp_parts(a_part, b_part)

                return -1
            else:
                if isinstance(b_part, int):
                    return 1

                return self._cmp_parts(a_part, b_part)

        return 0

    def __eq__(self, other):  # type: (Version) -> bool
        if not isinstance(other, Version):
            return NotImplemented

        return (
            self._major == other.major
            and self._minor == other.minor
            and self._patch == other.patch
            and self._prerelease == other.prerelease
            and self._build == other.build
        )

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return self._text

    def __repr__(self):
        return "<Version {}>".format(str(self))

    def __hash__(self):
        return hash(
            (
                self.major,
                self.minor,
                self.patch,
                ".".join(str(p) for p in self.prerelease),
                ".".join(str(p) for p in self.build),
            )
        )



import operator
import re

from .base_constraint import BaseConstraint
from .empty_constraint import EmptyConstraint
from .multi_constraint import MultiConstraint


class GenericConstraint(BaseConstraint):
    """
    Represents a generic constraint.

    This is particularly useful for platform/system/os/extra constraints.
    """

    OP_EQ = operator.eq
    OP_NE = operator.ne

    _trans_op_str = {"=": OP_EQ, "==": OP_EQ, "!=": OP_NE}

    _trans_op_int = {OP_EQ: "==", OP_NE: "!="}

    def __init__(self, operator, version):
        if operator not in self._trans_op_str:
            raise ValueError(
                'Invalid operator "{}" given, '
                "expected one of: {}".format(
                    operator, ", ".join(self.supported_operators)
                )
            )

        self._operator = self._trans_op_str[operator]
        self._string_operator = self._trans_op_int[self._operator]
        self._version = version

    @property
    def supported_operators(self):
        return list(self._trans_op_str.keys())

    @property
    def operator(self):
        return self._operator

    @property
    def string_operator(self):
        return self._string_operator

    @property
    def version(self):
        return self._version

    def matches(self, provider):
        if not isinstance(provider, GenericConstraint):
            return provider.matches(self)

        is_equal_op = self.OP_EQ is self._operator
        is_non_equal_op = self.OP_NE is self._operator
        is_provider_equal_op = self.OP_EQ is provider.operator
        is_provider_non_equal_op = self.OP_NE is provider.operator

        if (
            is_equal_op
            and is_provider_equal_op
            or is_non_equal_op
            and is_provider_non_equal_op
        ):
            return self._version == provider.version

        if (
            is_equal_op
            and is_provider_non_equal_op
            or is_non_equal_op
            and is_provider_equal_op
        ):
            return self._version != provider.version

        return False

    @classmethod
    def parse(cls, constraints):
        """
        Parses a constraint string into
        MultiConstraint and/or PlatformConstraint objects.
        """
        pretty_constraint = constraints

        or_constraints = re.split("\s*\|\|?\s*", constraints.strip())
        or_groups = []
        for constraints in or_constraints:
            and_constraints = re.split(
                "(?<!^)(?<![ ,]) *(?<!-)[, ](?!-) *(?!,|$)", constraints
            )
            if len(and_constraints) > 1:
                constraint_objects = []
                for constraint in and_constraints:
                    for parsed_constraint in cls._parse_constraint(constraint):
                        constraint_objects.append(parsed_constraint)
            else:
                constraint_objects = cls._parse_constraint(and_constraints[0])

            if len(constraint_objects) == 1:
                constraint = constraint_objects[0]
            else:
                constraint = MultiConstraint(constraint_objects)

            or_groups.append(constraint)

        if len(or_groups) == 1:
            constraint = or_groups[0]
        else:
            constraint = MultiConstraint(or_groups, False)

        constraint.pretty_string = pretty_constraint

        return constraint

    @classmethod
    def _parse_constraint(cls, constraint):
        m = re.match("(?i)^v?[xX*](\.[xX*])*$", constraint)
        if m:
            return (EmptyConstraint(),)

        # Basic Comparators
        m = re.match("^(!=|==?)?\s*(.*)", constraint)
        if m:
            return (GenericConstraint(m.group(1) or "=", m.group(2)),)

        raise ValueError("Could not parse generic constraint: {}".format(constraint))

    def __str__(self):
        op = self._trans_op_int[self._operator]
        if op == "==":
            op = ""
        else:
            op = op + " "

        return "{}{}".format(op, self._version)

    def __repr__(self):
        return "<GenericConstraint '{}'>".format(str(self))



from .. import tokens
from ..tokens import toml2py
from . import common
from .common import TokenElement
from .errors import InvalidElementError

_opening_bracket_types = (
    tokens.TYPE_OP_SQUARE_LEFT_BRACKET,
    tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET,
)
_closing_bracket_types = (
    tokens.TYPE_OP_SQUARE_RIGHT_BRACKET,
    tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET,
)
_name_types = (tokens.TYPE_BARE_STRING, tokens.TYPE_LITERAL_STRING, tokens.TYPE_STRING)


class TableHeaderElement(TokenElement):
    """
    An element containing opening and closing single and double square brackets, strings and dots and ending with
    a newline.
    
    Raises InvalidElementError.
    """

    def __init__(self, _tokens):
        TokenElement.__init__(self, _tokens, common.TYPE_MARKUP)
        self._names = tuple(
            toml2py.deserialize(token)
            for token in self._tokens
            if token.type in _name_types
        )

    @property
    def is_array_of_tables(self):
        opening_bracket = next(
            token
            for i, token in enumerate(self._tokens)
            if token.type in _opening_bracket_types
        )
        return opening_bracket.type == tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET

    @property
    def names(self):
        """
        Returns a sequence of string names making up this table header name.
        """
        return self._names

    def has_name_prefix(self, names):
        """
        Returns True if the header names is prefixed by the given sequence of names.
        """
        for i, name in enumerate(names):
            if self.names[i] != name:
                return False
        return True

    def serialized(self):
        return "".join(token.source_substring for token in self._tokens)

    def is_named(self, names):
        """
        Returns True if the given name sequence matches the full name of this header.
        """
        return tuple(names) == self.names

    def _validate_tokens(self, _tokens):

        opening_bracket_i = next(
            (
                i
                for i, token in enumerate(_tokens)
                if token.type in _opening_bracket_types
            ),
            float("-inf"),
        )

        if opening_bracket_i < 0:
            raise InvalidElementError("Expected an opening bracket")

        _tokens = _tokens[opening_bracket_i + 1 :]
        first_name_i = next(
            (i for i, token in enumerate(_tokens) if token.type in _name_types),
            float("-inf"),
        )
        if first_name_i < 0:
            raise InvalidElementError("Expected a table header name")

        _tokens = _tokens[first_name_i + 1 :]

        while True:

            next_dot_i = next(
                (
                    i
                    for i, token in enumerate(_tokens)
                    if token.type == tokens.TYPE_OPT_DOT
                ),
                float("-inf"),
            )
            if next_dot_i < 0:
                break

            _tokens = _tokens[next_dot_i + 1 :]

            next_name_i = next(
                (i for i, token in enumerate(_tokens) if token.type in _name_types),
                float("-inf"),
            )
            if next_name_i < 0:
                raise InvalidElementError("Expected a name after the dot")

            _tokens = _tokens[next_name_i + 1 :]

        closing_bracket_i = next(
            (
                i
                for i, token in enumerate(_tokens)
                if token.type in _closing_bracket_types
            ),
            float("-inf"),
        )

        if closing_bracket_i < 0:
            raise InvalidElementError("Expected a closing bracket")

        if _tokens[-1].type != tokens.TYPE_NEWLINE:
            raise InvalidElementError("Must end with a newline")



from typing import List

from poetry.mixology import resolve_version
from poetry.mixology.failure import SolveFailure
from poetry.packages.constraints.generic_constraint import GenericConstraint

from poetry.semver import parse_constraint

from .exceptions import CompatibilityError
from .exceptions import SolverProblemError

from .operations import Install
from .operations import Uninstall
from .operations import Update
from .operations.operation import Operation

from .provider import Provider


class Solver:
    def __init__(self, package, pool, installed, locked, io):
        self._package = package
        self._pool = pool
        self._installed = installed
        self._locked = locked
        self._io = io
        self._provider = Provider(self._package, self._pool, self._io)

    def solve(self, use_latest=None):  # type: (...) -> List[Operation]
        with self._provider.progress():
            packages, depths = self._solve(use_latest=use_latest)

        requested = self._package.all_requires

        operations = []
        for package in packages:
            installed = False
            for pkg in self._installed.packages:
                if package.name == pkg.name:
                    installed = True
                    # Checking version
                    if package.version != pkg.version:
                        operations.append(Update(pkg, package))
                    else:
                        operations.append(Install(package).skip("Already installed"))

                    break

            if not installed:
                operations.append(Install(package))

        # Checking for removals
        for pkg in self._locked.packages:
            remove = True
            for package in packages:
                if pkg.name == package.name:
                    remove = False
                    break

            if remove:
                skip = True
                for installed in self._installed.packages:
                    if installed.name == pkg.name:
                        skip = False
                        break

                op = Uninstall(pkg)
                if skip:
                    op.skip("Not currently installed")

                operations.append(op)

        return sorted(
            operations,
            key=lambda o: (
                o.job_type == "uninstall",
                # Packages to be uninstalled have no depth so we default to 0
                # since it actually doesn't matter since removals are always on top.
                -depths[packages.index(o.package)] if o.job_type != "uninstall" else 0,
                o.package.name,
            ),
        )

    def solve_in_compatibility_mode(self, constraints, use_latest=None):
        locked = {}
        for package in self._locked.packages:
            locked[package.name] = package

        packages = []
        depths = []
        for constraint in constraints:
            constraint = parse_constraint(constraint)
            intersection = constraint.intersect(self._package.python_constraint)

            self._provider.debug(
                "<comment>Retrying dependency resolution "
                "for Python ({}).</comment>".format(intersection)
            )
            with self._package.with_python_versions(str(intersection)):
                _packages, _depths = self._solve(use_latest=use_latest)
                for index, package in enumerate(_packages):
                    if package not in packages:
                        packages.append(package)
                        depths.append(_depths[index])
                        continue

                    current_package = packages[packages.index(package)]
                    for dep in package.requires:
                        if dep not in current_package.requires:
                            current_package.requires.append(dep)

        return packages, depths

    def _solve(self, use_latest=None):
        locked = {}
        for package in self._locked.packages:
            locked[package.name] = package

        try:
            result = resolve_version(
                self._package, self._provider, locked=locked, use_latest=use_latest
            )

            packages = result.packages
        except CompatibilityError as e:
            return self.solve_in_compatibility_mode(
                e.constraints, use_latest=use_latest
            )
        except SolveFailure as e:
            raise SolverProblemError(e)

        graph = self._build_graph(self._package, packages)

        depths = []
        for package in packages:
            category, optional, python, platform, depth = self._get_tags_for_package(
                package, graph
            )
            depths.append(depth)

            package.category = category
            package.optional = optional

            # If requirements are empty, drop them
            requirements = {}
            if python is not None and python != "*":
                requirements["python"] = python

            if platform is not None and platform != "*":
                requirements["platform"] = platform

            package.requirements = requirements

        return packages, depths

    def _build_graph(
        self, package, packages, previous=None, previous_dep=None, dep=None
    ):
        if not previous:
            category = "dev"
            optional = True
            python_version = "*"
            platform = "*"
        else:
            category = dep.category
            optional = dep.is_optional() and not dep.is_activated()
            python_version = str(
                parse_constraint(previous["python_version"]).intersect(
                    previous_dep.python_constraint
                )
            )
            platform = str(
                previous_dep.platform
                if GenericConstraint.parse(previous["platform"]).matches(
                    previous_dep.platform_constraint
                )
                and previous_dep.platform != "*"
                else previous["platform"]
            )

        graph = {
            "name": package.name,
            "category": category,
            "optional": optional,
            "python_version": python_version,
            "platform": platform,
            "children": [],
        }

        if previous_dep and previous_dep is not dep and previous_dep.name == dep.name:
            return graph

        for dependency in package.all_requires:
            if dependency.is_optional():
                if not package.is_root() and (
                    not previous_dep or not previous_dep.extras
                ):
                    continue

                is_activated = False
                for group, extras in package.extras.items():
                    if dep:
                        extras = previous_dep.extras
                    elif package.is_root():
                        extras = package.extras
                    else:
                        extras = []

                    if group in extras:
                        is_activated = True
                        break

                if not is_activated:
                    continue

            for pkg in packages:
                if pkg.name == dependency.name:
                    # If there is already a child with this name
                    # we merge the requirements
                    existing = None
                    for child in graph["children"]:
                        if (
                            child["name"] == pkg.name
                            and child["category"] == dependency.category
                        ):
                            existing = child
                            continue

                    child_graph = self._build_graph(
                        pkg, packages, graph, dependency, dep or dependency
                    )

                    if existing:
                        existing["python_version"] = str(
                            parse_constraint(existing["python_version"]).union(
                                parse_constraint(child_graph["python_version"])
                            )
                        )
                        continue

                    graph["children"].append(child_graph)

        return graph

    def _get_tags_for_package(self, package, graph, depth=0):
        categories = ["dev"]
        optionals = [True]
        python_versions = []
        platforms = []
        _depths = [0]

        children = graph["children"]
        found = False
        for child in children:
            if child["name"] == package.name:
                category = child["category"]
                optional = child["optional"]
                python_version = child["python_version"]
                platform = child["platform"]
                _depths.append(depth)
            else:
                (
                    category,
                    optional,
                    python_version,
                    platform,
                    _depth,
                ) = self._get_tags_for_package(package, child, depth=depth + 1)

                _depths.append(_depth)

            categories.append(category)
            optionals.append(optional)
            if python_version is not None:
                python_versions.append(python_version)

            if platform is not None:
                platforms.append(platform)

        if "main" in categories:
            category = "main"
        else:
            category = "dev"

        optional = all(optionals)

        if not python_versions:
            python_version = None
        else:
            # Find the least restrictive constraint
            python_version = python_versions[0]
            for constraint in python_versions[1:]:
                previous = parse_constraint(python_version)
                current = parse_constraint(constraint)

                if python_version == "*":
                    continue
                elif constraint == "*":
                    python_version = constraint
                elif current.allows_all(previous):
                    python_version = constraint

        if not platforms:
            platform = None
        else:
            platform = platforms[0]
            for constraint in platforms[1:]:
                previous = GenericConstraint.parse(platform)
                current = GenericConstraint.parse(constraint)

                if platform == "*":
                    continue
                elif constraint == "*":
                    platform = constraint
                elif current.matches(previous):
                    platform = constraint

        depth = max(*(_depths + [0]))

        return category, optional, python_version, platform, depth



class TOMLError(Exception):
    """
    All errors raised by this module are descendants of this type.
    """


class InvalidTOMLFileError(TOMLError):
    pass


class NoArrayFoundError(TOMLError):
    """
    An array of tables was requested but none exist by the given name.
    """


class InvalidValueError(TOMLError):
    pass


class DuplicateKeysError(TOMLError):
    """
    Duplicate keys detected in the parsed file.
    """


class DuplicateTablesError(TOMLError):
    """
    Duplicate tables detected in the parsed file.
    """



"""
PEP-517 compliant buildsystem API
"""
import logging
from pathlib import Path

from poetry.poetry import Poetry
from poetry.io import NullIO
from poetry.utils.venv import Venv

from .builders import SdistBuilder
from .builders import WheelBuilder

log = logging.getLogger(__name__)

# PEP 517 specifies that the CWD will always be the source tree
poetry = Poetry.create(".")


def get_requires_for_build_wheel(config_settings=None):
    """
    Returns a list of requirements for building, as strings
    """
    main, extras = SdistBuilder.convert_dependencies(poetry.package.requires)

    return main + extras


# For now, we require all dependencies to build either a wheel or an sdist.
get_requires_for_build_sdist = get_requires_for_build_wheel


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    """Builds a wheel, places it in wheel_directory"""
    info = WheelBuilder.make_in(poetry, NullIO(), Path(wheel_directory))

    return info.file.name


def build_sdist(sdist_directory, config_settings=None):
    """Builds an sdist, places it in sdist_directory"""
    path = SdistBuilder(poetry, Venv(), NullIO()).build(Path(sdist_directory))

    return path.name



import os
import shutil
import subprocess
import sys

from email.parser import Parser

from functools import cmp_to_key

from ..command import Command


class SelfUpdateCommand(Command):
    """
    Updates poetry to the latest version.

    self:update
        { version? : The version to update to. }
        { --preview : Install prereleases. }
    """

    def handle(self):
        from poetry.__version__ import __version__
        from poetry.repositories.pypi_repository import PyPiRepository
        from poetry.semver import Version
        from poetry.utils._compat import decode

        version = self.argument("version")
        if not version:
            version = ">=" + __version__

        repo = PyPiRepository(fallback=False)
        packages = repo.find_packages(
            "poetry", version, allow_prereleases=self.option("preview")
        )
        if not packages:
            self.line("No release found for the specified version")
            return

        packages.sort(
            key=cmp_to_key(
                lambda x, y: 0
                if x.version == y.version
                else int(x.version < y.version or -1)
            )
        )

        release = None
        for package in packages:
            if package.is_prerelease():
                if self.option("preview"):
                    release = package

                    break

                continue

            release = package

            break

        if release is None:
            self.line("No new release found")
            return

        if release.version == Version.parse(__version__):
            self.line("You are using the latest version")
            return

        try:
            self.update(release)
        except subprocess.CalledProcessError as e:
            self.line("")
            self.output.block(
                [
                    "[CalledProcessError]",
                    "An error has occured: {}".format(str(e)),
                    decode(e.output),
                ],
                style="error",
            )

            return e.returncode

    def update(self, release):
        from poetry.utils._compat import Path
        from poetry.utils.helpers import temporary_directory

        version = release.version
        self.line("Updating to <info>{}</info>".format(version))

        prefix = sys.prefix
        base_prefix = getattr(sys, "base_prefix", None)
        real_prefix = getattr(sys, "real_prefix", None)

        prefix_poetry = self._bin_path(Path(prefix), "poetry")
        if prefix_poetry.exists():
            pip = self._bin_path(prefix_poetry.parent.parent, "pip").resolve()
        elif (
            base_prefix
            and base_prefix != prefix
            and self._bin_path(Path(base_prefix), "poetry").exists()
        ):
            pip = self._bin_path(Path(base_prefix), "pip")
        elif real_prefix:
            pip = self._bin_path(Path(real_prefix), "pip")
        else:
            pip = self._bin_path(Path(prefix), "pip")

            if not pip.exists():
                raise RuntimeError("Unable to determine poetry's path")

        with temporary_directory(prefix="poetry-update-") as temp_dir:
            temp_dir = Path(temp_dir)
            dist = temp_dir / "dist"
            self.line("  - Getting dependencies")
            self.process(
                str(pip),
                "install",
                "-U",
                "poetry=={}".format(release.version),
                "--target",
                str(dist),
            )

            self.line("  - Vendorizing dependencies")

            poetry_dir = dist / "poetry"
            vendor_dir = poetry_dir / "_vendor"

            # Everything, except poetry itself, should
            # be put in the _vendor directory
            for file in dist.glob("*"):
                if file.name.startswith("poetry"):
                    continue

                dest = vendor_dir / file.name
                if file.is_dir():
                    shutil.copytree(str(file), str(dest))
                    shutil.rmtree(str(file))
                else:
                    shutil.copy(str(file), str(dest))
                    os.unlink(str(file))

            wheel_data = dist / "poetry-{}.dist-info".format(version) / "WHEEL"
            with wheel_data.open() as f:
                wheel_data = Parser().parsestr(f.read())

            tag = wheel_data["Tag"]

            # Repack everything and install
            self.line("  - Updating <info>poetry</info>")

            shutil.make_archive(
                str(temp_dir / "poetry-{}-{}".format(version, tag)),
                format="zip",
                root_dir=str(dist),
            )

            os.rename(
                str(temp_dir / "poetry-{}-{}.zip".format(version, tag)),
                str(temp_dir / "poetry-{}-{}.whl".format(version, tag)),
            )

            self.process(
                str(pip),
                "install",
                "--upgrade",
                "--no-deps",
                str(temp_dir / "poetry-{}-{}.whl".format(version, tag)),
            )

            self.line("")
            self.line(
                "<info>poetry</> (<comment>{}</>) "
                "successfully installed!".format(version)
            )

    def process(self, *args):
        return subprocess.check_output(list(args), stderr=subprocess.STDOUT)

    def _bin_path(self, base_path, bin):
        if sys.platform == "win32":
            return (base_path / "Scripts" / bin).with_suffix(".exe")

        return base_path / "bin" / bin




"""
     The following predicates can be used in the traversal functions directly.
"""

from ...tokens import TYPE_OP_ASSIGNMENT
from ...tokens import TYPE_OP_COMMA
from ...tokens import TYPE_OP_SQUARE_LEFT_BRACKET
from ...tokens import TYPE_OP_SQUARE_RIGHT_BRACKET
from ..atomic import AtomicElement
from ..metadata import (
    PunctuationElement,
    CommentElement,
    NewlineElement,
    WhitespaceElement,
)
from .. import common


atomic = lambda e: isinstance(e, AtomicElement)


op_assignment = (
    lambda e: isinstance(e, PunctuationElement) and e.token.type == TYPE_OP_ASSIGNMENT
)


op_comma = lambda e: isinstance(e, PunctuationElement) and e.token.type == TYPE_OP_COMMA


comment = lambda e: isinstance(e, CommentElement)


newline = lambda e: isinstance(e, NewlineElement)


non_metadata = lambda e: e.type != common.TYPE_METADATA


closing_square_bracket = (
    lambda e: isinstance(e, PunctuationElement)
    and e.token.type == TYPE_OP_SQUARE_RIGHT_BRACKET
)


opening_square_bracket = (
    lambda e: isinstance(e, PunctuationElement)
    and e.token.type == TYPE_OP_SQUARE_LEFT_BRACKET
)


def table(e):
    from ..table import TableElement

    return isinstance(e, TableElement)


def table_header(e):
    from ..tableheader import TableHeaderElement

    return isinstance(e, TableHeaderElement)


whitespace = lambda e: isinstance(e, WhitespaceElement)



# -*- coding: utf-8 -*-

from typing import Union


class Operation(object):
    def __init__(self, reason=None):  # type: (Union[str, None]) -> None
        self._reason = reason

        self._skipped = False
        self._skip_reason = None

    @property
    def job_type(self):  # type: () -> str
        raise NotImplementedError

    @property
    def reason(self):  # type: () -> str
        return self._reason

    @property
    def skipped(self):  # type: () -> bool
        return self._skipped

    @property
    def skip_reason(self):  # type: () -> Union[str, None]
        return self._skip_reason

    def format_version(self, package):  # type: (...) -> str
        return package.full_pretty_version

    def skip(self, reason):  # type: (str) -> Operation
        self._skipped = True
        self._skip_reason = reason

        return self

    def unskip(self):  # type: () -> Operation
        self._skipped = False
        self._skip_reason = None

        return self



import os
import tempfile

from subprocess import CalledProcessError

try:
    import urllib.parse as urlparse
except ImportError:
    import urlparse

from poetry.utils._compat import encode
from poetry.utils.venv import Venv

from .base_installer import BaseInstaller


class PipInstaller(BaseInstaller):
    def __init__(self, venv, io):  # type: (Venv, ...) -> None
        self._venv = venv
        self._io = io

    def install(self, package, update=False):
        args = ["install", "--no-deps"]

        if package.source_type == "legacy" and package.source_url:
            parsed = urlparse.urlparse(package.source_url)
            if parsed.scheme == "http":
                self._io.write_error(
                    "    <warning>Installing from unsecure host: {}</warning>".format(
                        parsed.netloc
                    )
                )
                args += ["--trusted-host", parsed.netloc]

            args += ["--index-url", package.source_url]

        if update:
            args.append("-U")

        if package.hashes and not package.source_type:
            # Format as a requirements.txt
            # We need to create a requirements.txt file
            # for each package in order to check hashes.
            # This is far from optimal but we do not have any
            # other choice since this is the only way for pip
            # to verify hashes.
            req = self.create_temporary_requirement(package)
            args += ["-r", req]

            try:
                self.run(*args)
            finally:
                os.unlink(req)
        else:
            req = self.requirement(package)
            if not isinstance(req, list):
                args.append(req)
            else:
                args += req

            self.run(*args)

    def update(self, _, target):
        self.install(target, update=True)

    def remove(self, package):
        try:
            self.run("uninstall", package.name, "-y")
        except CalledProcessError as e:
            if "not installed" in str(e):
                return

            raise

    def run(self, *args, **kwargs):  # type: (...) -> str
        return self._venv.run("pip", *args, **kwargs)

    def requirement(self, package, formatted=False):
        if formatted and not package.source_type:
            req = "{}=={}".format(package.name, package.version)
            for h in package.hashes:
                req += " --hash sha256:{}".format(h)

            req += "\n"

            return req

        if package.source_type in ["file", "directory"]:
            if package.root_dir:
                req = os.path.join(package.root_dir, package.source_url)
            else:
                req = os.path.realpath(package.source_url)

            if package.develop:
                req = ["-e", req]

            return req

        if package.source_type == "git":
            return "git+{}@{}#egg={}".format(
                package.source_url, package.source_reference, package.name
            )

        return "{}=={}".format(package.name, package.version)

    def create_temporary_requirement(self, package):
        fd, name = tempfile.mkstemp(
            "reqs.txt", "{}-{}".format(package.name, package.version)
        )

        try:
            os.write(fd, encode(self.requirement(package, formatted=True)))
        finally:
            os.close(fd)

        return name



from .package import Package


class ProjectPackage(Package):
    def __init__(self, name, version, pretty_version=None):
        super(ProjectPackage, self).__init__(name, version, pretty_version)

        self.build = None
        self.packages = []
        self.include = []
        self.exclude = []

    def is_root(self):
        return True

    def to_dependency(self):
        dependency = super(ProjectPackage, self).to_dependency()

        dependency.is_root = True

        return dependency



from os import environ
from distutils.util import strtobool

from .venv_command import VenvCommand


class ShellCommand(VenvCommand):
    """
    Spawns a shell within the virtual environment.

    shell [options]
    """

    help = """The <info>shell</> command spawns a shell, according to the
<comment>$SHELL</> environment variable, within the virtual environment.
If one doesn't exist yet, it will be created.
"""

    def handle(self):
        from poetry.utils.shell import Shell

        # Check if it's already activated or doesn't exist and won't be created
        if strtobool(environ.get("POETRY_ACTIVE", "0")) or not self.venv.is_venv():
            current_venv = environ.get("VIRTUAL_ENV")

            if current_venv:
                self.line(
                    "Virtual environment already activated: "
                    "<info>{}</>".format(current_venv)
                )

            else:
                self.error("Virtual environment wasn't found")

            return

        self.line("Spawning shell within <info>{}</>".format(self.venv.venv))

        # Setting this to avoid spawning unnecessary nested shells
        environ["POETRY_ACTIVE"] = "1"
        shell = Shell.get()
        self.venv.execute(shell.path)
        environ.pop("POETRY_ACTIVE")



from .prettify.errors import InvalidValueError
from .freshtable import FreshTable
from .prettify import util


class ArrayOfTables(list):
    def __init__(self, toml_file, name, iterable=None):
        if iterable:
            list.__init__(self, iterable)
        self._name = name
        self._toml_file = toml_file

    def append(self, value):
        if isinstance(value, dict):
            table = FreshTable(parent=self, name=self._name, is_array=True)
            table._append_to_parent()
            index = len(self._toml_file[self._name]) - 1
            for key_seq, value in util.flatten_nested(value).items():
                # self._toml_file._setitem_with_key_seq((self._name, index) + key_seq, value)
                self._toml_file._array_setitem_with_key_seq(
                    self._name, index, key_seq, value
                )
            # for k, v in value.items():
            #     table[k] = v
        else:
            raise InvalidValueError("Can only append a dict to an array of tables")

    def __getitem__(self, item):
        try:
            return list.__getitem__(self, item)
        except IndexError:
            if item == len(self):
                return FreshTable(parent=self, name=self._name, is_array=True)
            else:
                raise

    def append_fresh_table(self, fresh_table):
        list.append(self, fresh_table)
        if self._toml_file:
            self._toml_file.append_fresh_table(fresh_table)



from .command import Command


class VenvCommand(Command):
    def __init__(self):
        self._venv = None

        super(VenvCommand, self).__init__()

    def initialize(self, i, o):
        from poetry.utils.venv import Venv

        super(VenvCommand, self).initialize(i, o)

        self._venv = Venv.create(
            o, self.poetry.package.name, cwd=self.poetry.file.parent
        )

        if self._venv.is_venv() and o.is_verbose():
            o.writeln("Using virtualenv: <comment>{}</>".format(self._venv.venv))

    @property
    def venv(self):
        return self._venv



import datetime
import six

from .. import tokens
from ..tokens import py2toml
from ..util import join_with
from .atomic import AtomicElement
from .metadata import PunctuationElement, WhitespaceElement, NewlineElement
from .tableheader import TableHeaderElement


def create_element(value, multiline_strings_allowed=True):
    """
    Creates and returns the appropriate elements.Element instance from the given Python primitive, sequence-like,
    or dict-like value.
    """
    from .array import ArrayElement

    if (
        isinstance(
            value,
            (int, float, bool, datetime.datetime, datetime.date) + six.string_types,
        )
        or value is None
    ):
        primitive_token = py2toml.create_primitive_token(
            value, multiline_strings_allowed=multiline_strings_allowed
        )
        return AtomicElement((primitive_token,))

    elif isinstance(value, (list, tuple)):
        preamble = [create_operator_element("[")]
        postable = [create_operator_element("]")]
        stuffing_elements = [create_element(v) for v in value]
        spaced_stuffing = join_with(
            stuffing_elements,
            separator=[create_operator_element(","), create_whitespace_element()],
        )

        return ArrayElement(preamble + spaced_stuffing + postable)

    elif isinstance(value, dict):
        return create_inline_table(
            value,
            multiline_table=False,
            multiline_strings_allowed=multiline_strings_allowed,
        )

    else:
        raise RuntimeError(
            "Value type unaccounted for: {} of type {}".format(value, type(value))
        )


def create_inline_table(
    from_dict, multiline_table=False, multiline_strings_allowed=True
):
    """
    Creates an InlineTable element from the given dict instance.
    """

    from .inlinetable import InlineTableElement

    preamble = [create_operator_element("{")]
    postable = [create_operator_element("}")]

    stuffing_elements = (
        (
            create_string_element(k, bare_allowed=True),
            create_whitespace_element(),
            create_operator_element("="),
            create_whitespace_element(),
            create_element(v, multiline_strings_allowed=False),
        )
        for (k, v) in from_dict.items()
    )

    pair_separator = [
        create_operator_element(","),
        create_newline_element() if multiline_table else create_whitespace_element(),
    ]
    spaced_elements = join_with(stuffing_elements, separator=pair_separator)

    return InlineTableElement(preamble + spaced_elements + postable)


def create_string_element(value, bare_allowed=False):
    """
    Creates and returns an AtomicElement wrapping a string value.
    """
    return AtomicElement((py2toml.create_string_token(value, bare_allowed),))


def create_operator_element(operator):
    """
    Creates a PunctuationElement instance containing an operator token of the specified type. The operator
    should be a TOML source str.
    """
    operator_type_map = {
        ",": tokens.TYPE_OP_COMMA,
        "=": tokens.TYPE_OP_ASSIGNMENT,
        "[": tokens.TYPE_OP_SQUARE_LEFT_BRACKET,
        "]": tokens.TYPE_OP_SQUARE_RIGHT_BRACKET,
        "[[": tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET,
        "]]": tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET,
        "{": tokens.TYPE_OP_CURLY_LEFT_BRACKET,
        "}": tokens.TYPE_OP_CURLY_RIGHT_BRACKET,
    }

    ts = (tokens.Token(operator_type_map[operator], operator),)
    return PunctuationElement(ts)


def create_newline_element():
    """
    Creates and returns a single NewlineElement.
    """
    ts = (tokens.Token(tokens.TYPE_NEWLINE, "\n"),)
    return NewlineElement(ts)


def create_whitespace_element(length=1, char=" "):
    """
    Creates and returns a WhitespaceElement containing spaces.
    """
    ts = (tokens.Token(tokens.TYPE_WHITESPACE, char),) * length
    return WhitespaceElement(ts)


def create_table_header_element(names):

    name_elements = []

    if isinstance(names, six.string_types):
        name_elements = [py2toml.create_string_token(names, bare_string_allowed=True)]
    else:
        for (i, name) in enumerate(names):
            name_elements.append(
                py2toml.create_string_token(name, bare_string_allowed=True)
            )
            if i < (len(names) - 1):
                name_elements.append(py2toml.operator_token(tokens.TYPE_OPT_DOT))

    return TableHeaderElement(
        [py2toml.operator_token(tokens.TYPE_OP_SQUARE_LEFT_BRACKET)]
        + name_elements
        + [
            py2toml.operator_token(tokens.TYPE_OP_SQUARE_RIGHT_BRACKET),
            py2toml.operator_token(tokens.TYPE_NEWLINE),
        ]
    )


def create_array_of_tables_header_element(name):
    return TableHeaderElement(
        (
            py2toml.operator_token(tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET),
            py2toml.create_string_token(name, bare_string_allowed=True),
            py2toml.operator_token(tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET),
            py2toml.operator_token(tokens.TYPE_NEWLINE),
        )
    )


def create_table(dict_value):
    """
    Creates a TableElement out of a dict instance.
    """
    from .table import TableElement

    if not isinstance(dict_value, dict):
        raise ValueError("input must be a dict instance.")

    table_element = TableElement([create_newline_element()])
    for k, v in dict_value.items():
        table_element[k] = create_element(v)

    return table_element


def create_multiline_string(text, maximum_line_length):
    return AtomicElement(
        _tokens=[py2toml.create_multiline_string(text, maximum_line_length)]
    )



from __future__ import absolute_import
from __future__ import unicode_literals

import json

import jsonschema

from .__version__ import __version__
from .config import Config
from .exceptions import InvalidProjectFile
from .packages import Dependency
from .packages import Locker
from .packages import Package
from .packages import ProjectPackage
from .repositories import Pool
from .repositories.pypi_repository import PyPiRepository
from .spdx import license_by_id
from .utils._compat import Path
from .utils.toml_file import TomlFile


class Poetry:

    VERSION = __version__

    def __init__(
        self,
        file,  # type: Path
        local_config,  # type: dict
        package,  # type: Package
        locker,  # type: Locker
    ):
        self._file = TomlFile(file)
        self._package = package
        self._local_config = local_config
        self._locker = locker
        self._config = Config.create("config.toml")

        # Configure sources
        self._pool = Pool()
        for source in self._local_config.get("source", []):
            self._pool.configure(source)

        # Always put PyPI last to prefere private repositories
        self._pool.add_repository(PyPiRepository())

    @property
    def file(self):
        return self._file

    @property
    def package(self):  # type: () -> Package
        return self._package

    @property
    def local_config(self):  # type: () -> dict
        return self._local_config

    @property
    def locker(self):  # type: () -> Locker
        return self._locker

    @property
    def pool(self):  # type: () -> Pool
        return self._pool

    @classmethod
    def create(cls, cwd):  # type: (Path) -> Poetry
        candidates = [Path(cwd)]
        candidates.extend(Path(cwd).parents)

        for path in candidates:
            poetry_file = path / "pyproject.toml"

            if poetry_file.exists():
                break

        else:
            raise RuntimeError(
                "Poetry could not find a pyproject.toml file in {} or its parents".format(
                    cwd
                )
            )

        local_config = TomlFile(poetry_file.as_posix()).read(True)
        if "tool" not in local_config or "poetry" not in local_config["tool"]:
            raise RuntimeError(
                "[tool.poetry] section not found in {}".format(poetry_file.name)
            )
        local_config = local_config["tool"]["poetry"]

        # Checking validity
        cls.check(local_config)

        # Load package
        name = local_config["name"]
        version = local_config["version"]
        package = ProjectPackage(name, version, version)
        package.root_dir = poetry_file.parent

        for author in local_config["authors"]:
            package.authors.append(author)

        package.description = local_config.get("description", "")
        package.homepage = local_config.get("homepage")
        package.repository_url = local_config.get("repository")
        package.license = local_config.get("license")
        package.keywords = local_config.get("keywords", [])
        package.classifiers = local_config.get("classifiers", [])

        if "readme" in local_config:
            package.readme = Path(poetry_file.parent) / local_config["readme"]

        if "platform" in local_config:
            package.platform = local_config["platform"]

        if "dependencies" in local_config:
            for name, constraint in local_config["dependencies"].items():
                if name.lower() == "python":
                    package.python_versions = constraint
                    continue

                package.add_dependency(name, constraint)

        if "dev-dependencies" in local_config:
            for name, constraint in local_config["dev-dependencies"].items():
                package.add_dependency(name, constraint, category="dev")

        extras = local_config.get("extras", {})
        for extra_name, requirements in extras.items():
            package.extras[extra_name] = []

            # Checking for dependency
            for req in requirements:
                req = Dependency(req, "*")

                for dep in package.requires:
                    if dep.name == req.name:
                        dep.in_extras.append(extra_name)
                        package.extras[extra_name].append(dep)

                        break

        if "build" in local_config:
            package.build = local_config["build"]

        if "include" in local_config:
            package.include = local_config["include"]

        if "exclude" in local_config:
            package.exclude = local_config["exclude"]

        if "packages" in local_config:
            package.packages = local_config["packages"]

        locker = Locker(poetry_file.with_suffix(".lock"), local_config)

        return cls(poetry_file, local_config, package, locker)

    @classmethod
    def check(cls, config, strict=False):  # type: (dict, bool) -> bool
        """
        Checks the validity of a configuration
        """
        schema = Path(__file__).parent / "json" / "schemas" / "poetry-schema.json"

        with schema.open() as f:
            schema = json.loads(f.read())

        try:
            jsonschema.validate(config, schema)
        except jsonschema.ValidationError as e:
            message = e.message
            if e.path:
                message = "[{}] {}".format(".".join(e.path), message)

            raise InvalidProjectFile(message)

        if strict:
            # If strict, check the file more thoroughly

            # Checking license
            license = config.get("license")
            if license:
                try:
                    license_by_id(license)
                except ValueError:
                    raise InvalidProjectFile("Invalid license")

        return True



from .venv_command import VenvCommand


class UpdateCommand(VenvCommand):
    """
    Update dependencies as according to the <comment>pyproject.toml</> file.

    update
        { packages?* : The packages to update. }
        { --no-dev : Do not install dev dependencies. }
        { --dry-run : Outputs the operations but will not execute anything
                      (implicitly enables --verbose). }
    """

    _loggers = ["poetry.repositories.pypi_repository"]

    def handle(self):
        from poetry.installation import Installer

        packages = self.argument("packages")

        installer = Installer(
            self.output,
            self.venv,
            self.poetry.package,
            self.poetry.locker,
            self.poetry.pool,
        )

        if packages:
            installer.whitelist({name: "*" for name in packages})

        installer.dev_mode(not self.option("no-dev"))
        installer.dry_run(self.option("dry-run"))

        # Force update
        installer.update(True)

        return installer.run()



from typing import List
from typing import Union

import poetry.packages

from .base_repository import BaseRepository
from .repository import Repository


class Pool(BaseRepository):
    def __init__(self, repositories=None):  # type: (Union[list, None]) -> None
        if repositories is None:
            repositories = []

        self._repositories = []

        for repository in repositories:
            self.add_repository(repository)

        super(Pool, self).__init__()

    @property
    def repositories(self):  # type: () -> List[Repository]
        return self._repositories

    def add_repository(self, repository):  # type: (Repository) -> Pool
        """
        Adds a repository to the pool.
        """
        self._repositories.append(repository)

        return self

    def remove_repository(self, repository_name):  # type: (str) -> Pool
        for i, repository in enumerate(self._repositories):
            if repository.name == repository_name:
                del self._repositories[i]

                break

        return self

    def configure(self, source):  # type: (dict) -> Pool
        """
        Configures a repository based on a source
        specification and add it to the pool.
        """
        from .legacy_repository import LegacyRepository

        if "url" in source:
            # PyPI-like repository
            if "name" not in source:
                raise RuntimeError("Missing [name] in source.")

            repository = LegacyRepository(source["name"], source["url"])
        else:
            raise RuntimeError("Unsupported source specified")

        return self.add_repository(repository)

    def has_package(self, package):
        raise NotImplementedError()

    def package(self, name, version, extras=None):
        for repository in self._repositories:
            try:
                package = repository.package(name, version, extras=extras)
            except ValueError:
                continue

            if package:
                self._packages.append(package)

                return package

        return None

    def find_packages(
        self, name, constraint=None, extras=None, allow_prereleases=False
    ):
        for repository in self._repositories:
            packages = repository.find_packages(
                name, constraint, extras=extras, allow_prereleases=allow_prereleases
            )
            if packages:
                return packages

        return []

    def search(self, query, mode=BaseRepository.SEARCH_FULLTEXT):
        from .legacy_repository import LegacyRepository

        results = []
        for repository in self._repositories:
            if isinstance(repository, LegacyRepository):
                continue

            results += repository.search(query, mode=mode)

        return results



import operator

from pyparsing import ParseException, ParseResults, stringStart, stringEnd

from pyparsing import ZeroOrMore, Group, Forward, QuotedString
from pyparsing import Literal as L  # noqa


class InvalidMarker(ValueError):
    """
    An invalid marker was found, users should refer to PEP 508.
    """


class UndefinedComparison(ValueError):
    """
    An invalid operation was attempted on a value that doesn't support it.
    """


class UndefinedEnvironmentName(ValueError):
    """
    A name was attempted to be used that does not exist inside of the
    environment.
    """


class Node(object):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return "<{0}({1!r})>".format(self.__class__.__name__, str(self))

    def serialize(self):
        raise NotImplementedError


class Variable(Node):
    def serialize(self):
        return str(self)


class Value(Node):
    def serialize(self):
        return '"{0}"'.format(self)


class Op(Node):
    def serialize(self):
        return str(self)


VARIABLE = (
    L("implementation_version")
    | L("platform_python_implementation")
    | L("implementation_name")
    | L("python_full_version")
    | L("platform_release")
    | L("platform_version")
    | L("platform_machine")
    | L("platform_system")
    | L("python_version")
    | L("sys_platform")
    | L("os_name")
    | L("os.name")
    | L("sys.platform")  # PEP-345
    | L("platform.version")  # PEP-345
    | L("platform.machine")  # PEP-345
    | L("platform.python_implementation")  # PEP-345
    | L("python_implementation")  # PEP-345
    | L("extra")  # undocumented setuptools legacy
)
ALIASES = {
    "os.name": "os_name",
    "sys.platform": "sys_platform",
    "platform.version": "platform_version",
    "platform.machine": "platform_machine",
    "platform.python_implementation": "platform_python_implementation",
    "python_implementation": "platform_python_implementation",
}
VARIABLE.setParseAction(lambda s, l, t: Variable(ALIASES.get(t[0], t[0])))

VERSION_CMP = (
    L("===") | L("==") | L(">=") | L("<=") | L("!=") | L("~=") | L(">") | L("<")
)

MARKER_OP = VERSION_CMP | L("not in") | L("in")
MARKER_OP.setParseAction(lambda s, l, t: Op(t[0]))

MARKER_VALUE = QuotedString("'") | QuotedString('"')
MARKER_VALUE.setParseAction(lambda s, l, t: Value(t[0]))

BOOLOP = L("and") | L("or")

MARKER_VAR = VARIABLE | MARKER_VALUE

MARKER_ITEM = Group(MARKER_VAR + MARKER_OP + MARKER_VAR)
MARKER_ITEM.setParseAction(lambda s, l, t: tuple(t[0]))

LPAREN = L("(").suppress()
RPAREN = L(")").suppress()

MARKER_EXPR = Forward()
MARKER_ATOM = MARKER_ITEM | Group(LPAREN + MARKER_EXPR + RPAREN)
MARKER_EXPR << MARKER_ATOM + ZeroOrMore(BOOLOP + MARKER_EXPR)

MARKER = stringStart + MARKER_EXPR + stringEnd


def _coerce_parse_result(results):
    if isinstance(results, ParseResults):
        return [_coerce_parse_result(i) for i in results]
    else:
        return results


def _format_marker(marker, first=True):
    assert isinstance(marker, (list, tuple, str))

    # Sometimes we have a structure like [[...]] which is a single item list
    # where the single item is itself it's own list. In that case we want skip
    # the rest of this function so that we don't get extraneous () on the
    # outside.
    if (
        isinstance(marker, list)
        and len(marker) == 1
        and isinstance(marker[0], (list, tuple))
    ):
        return _format_marker(marker[0])

    if isinstance(marker, list):
        inner = (_format_marker(m, first=False) for m in marker)
        if first:
            return " ".join(inner)
        else:
            return "(" + " ".join(inner) + ")"
    elif isinstance(marker, tuple):
        return " ".join([m.serialize() for m in marker])
    else:
        return marker


_operators = {
    "in": lambda lhs, rhs: lhs in rhs,
    "not in": lambda lhs, rhs: lhs not in rhs,
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    ">": operator.gt,
}


def format_full_version(info):
    version = "{0.major}.{0.minor}.{0.micro}".format(info)
    kind = info.releaselevel
    if kind != "final":
        version += kind[0] + str(info.serial)
    return version


class Marker(object):
    def __init__(self, marker):
        try:
            self._markers = _coerce_parse_result(MARKER.parseString(marker))
        except ParseException as e:
            err_str = "Invalid marker: {0!r}, parse error at {1!r}".format(
                marker, marker[e.loc : e.loc + 8]
            )
            raise InvalidMarker(err_str)

    @property
    def markers(self):
        return self._markers

    def __str__(self):
        return _format_marker(self._markers)

    def __repr__(self):
        return "<Marker({0!r})>".format(str(self))



import os
import platform
import subprocess
import sys
import sysconfig
import warnings

from contextlib import contextmanager
from subprocess import CalledProcessError

from poetry.config import Config
from poetry.locations import CACHE_DIR
from poetry.utils._compat import Path
from poetry.utils._compat import decode


class VenvError(Exception):

    pass


class VenvCommandError(VenvError):
    def __init__(self, e):  # type: (CalledProcessError) -> None
        message = "Command {} errored with the following output: \n{}".format(
            e.cmd, decode(e.output)
        )

        super(VenvCommandError, self).__init__(message)


class Venv(object):
    def __init__(self, venv=None):
        self._venv = venv
        if self._venv:
            self._venv = Path(self._venv)

        self._windows = sys.platform == "win32"

        self._bin_dir = None
        if venv:
            bin_dir = "bin" if not self._windows else "Scripts"
            self._bin_dir = self._venv / bin_dir

        self._version_info = None
        self._python_implementation = None

    @classmethod
    def create(cls, io, name=None, cwd=None):  # type: (...) -> Venv
        if "VIRTUAL_ENV" not in os.environ:
            # Not in a virtualenv
            # Checking if we need to create one

            # First we check if there is a .venv
            # at the root of the project.
            if cwd and (cwd / ".venv").exists():
                venv = cwd / ".venv"
            else:
                config = Config.create("config.toml")

                create_venv = config.setting("settings.virtualenvs.create")
                root_venv = config.setting("settings.virtualenvs.in-project")

                venv_path = config.setting("settings.virtualenvs.path")
                if root_venv:
                    if not cwd:
                        raise RuntimeError(
                            "Unbale to determine the project's directory"
                        )

                    venv_path = cwd / ".venv"
                elif venv_path is None:
                    venv_path = Path(CACHE_DIR) / "virtualenvs"
                else:
                    venv_path = Path(venv_path)

                if not name:
                    name = Path.cwd().name

                name = "{}-py{}".format(
                    name, ".".join([str(v) for v in sys.version_info[:2]])
                )

                if root_venv:
                    venv = venv_path
                else:
                    venv = venv_path / name

                if not venv.exists():
                    if create_venv is False:
                        io.writeln(
                            "<fg=black;bg=yellow>"
                            "Skipping virtualenv creation, "
                            "as specified in config file."
                            "</>"
                        )

                        return cls()

                    io.writeln(
                        "Creating virtualenv <info>{}</> in {}".format(
                            name, str(venv_path)
                        )
                    )

                    cls.build(str(venv))
                else:
                    if io.is_very_verbose():
                        io.writeln(
                            "Virtualenv <info>{}</> already exists.".format(name)
                        )

            os.environ["VIRTUAL_ENV"] = str(venv)

        # venv detection:
        # stdlib venv may symlink sys.executable, so we can't use realpath.
        # but others can symlink *to* the venv Python,
        # so we can't just use sys.executable.
        # So we just check every item in the symlink tree (generally <= 3)
        p = os.path.normcase(sys.executable)
        paths = [p]
        while os.path.islink(p):
            p = os.path.normcase(os.path.join(os.path.dirname(p), os.readlink(p)))
            paths.append(p)

        p_venv = os.path.normcase(os.environ["VIRTUAL_ENV"])
        if any(p.startswith(p_venv) for p in paths):
            # Running properly in the virtualenv, don't need to do anything
            return cls()

        venv = os.environ["VIRTUAL_ENV"]

        return cls(venv)

    @classmethod
    def build(cls, path):
        try:
            from venv import EnvBuilder

            builder = EnvBuilder(with_pip=True)
            build = builder.create
        except ImportError:
            # We fallback on virtualenv for Python 2.7
            from virtualenv import create_environment

            build = create_environment

        build(path)

    @property
    def venv(self):
        return self._venv

    @property
    def python(self):  # type: () -> str
        """
        Path to current python executable
        """
        return self._bin("python")

    @property
    def pip(self):  # type: () -> str
        """
        Path to current pip executable
        """
        return self._bin("pip")

    @property
    def version_info(self):  # type: () -> tuple
        if self._version_info is not None:
            return self._version_info

        if not self.is_venv():
            self._version_info = sys.version_info
        else:
            output = self.run(
                "python",
                "-c",
                "\"import sys; print('.'.join([str(s) for s in sys.version_info[:3]]))\"",
                shell=True,
            )

            self._version_info = tuple([int(s) for s in output.strip().split(".")])

        return self._version_info

    @property
    def python_implementation(self):
        if self._python_implementation is not None:
            return self._python_implementation

        if not self.is_venv():
            impl = platform.python_implementation()
        else:
            impl = self.run(
                "python",
                "-c",
                '"import platform; print(platform.python_implementation())"',
                shell=True,
            ).strip()

        self._python_implementation = impl

        return self._python_implementation

    def config_var(self, var):
        if not self.is_venv():
            try:
                return sysconfig.get_config_var(var)
            except IOError as e:
                warnings.warn("{0}".format(e), RuntimeWarning)
                return None

        try:
            value = self.run(
                "python",
                "-c",
                '"import sysconfig; '
                "print(sysconfig.get_config_var('{}'))\"".format(var),
                shell=True,
            ).strip()
        except VenvCommandError as e:
            warnings.warn("{0}".format(e), RuntimeWarning)
            return None

        if value == "None":
            value = None
        elif value == "1":
            value = 1
        elif value == "0":
            value = 0

        return value

    def run(self, bin, *args, **kwargs):
        """
        Run a command inside the virtual env.
        """
        if self._windows:
            bin = self._bin(bin)

        cmd = [bin] + list(args)
        shell = kwargs.get("shell", False)
        call = kwargs.pop("call", False)

        if shell:
            cmd = " ".join(cmd)

        try:
            if not self.is_venv():
                if call:
                    return subprocess.call(cmd, stderr=subprocess.STDOUT, **kwargs)

                output = subprocess.check_output(
                    cmd, stderr=subprocess.STDOUT, **kwargs
                )
            else:
                if self._windows:
                    kwargs["shell"] = True

                with self.temp_environ():
                    os.environ["PATH"] = self._path()
                    os.environ["VIRTUAL_ENV"] = str(self._venv)

                    self.unset_env("PYTHONHOME")
                    self.unset_env("__PYVENV_LAUNCHER__")

                    if call:
                        return subprocess.call(cmd, stderr=subprocess.STDOUT, **kwargs)

                    output = subprocess.check_output(
                        cmd, stderr=subprocess.STDOUT, **kwargs
                    )
        except CalledProcessError as e:
            raise VenvCommandError(e)

        return decode(output)

    def execute(self, bin, *args, **kwargs):
        if not self.is_venv():
            return subprocess.call([bin] + list(args))
        else:
            if self._windows:
                bin = self._bin(bin)

            with self.temp_environ():
                os.environ["PATH"] = self._path()
                os.environ["VIRTUAL_ENV"] = str(self._venv)

                self.unset_env("PYTHONHOME")
                self.unset_env("__PYVENV_LAUNCHER__")

                return subprocess.call([bin] + list(args), **kwargs)

    @contextmanager
    def temp_environ(self):
        environ = dict(os.environ)
        try:
            yield
        finally:
            os.environ.clear()
            os.environ.update(environ)

    def _path(self):
        return os.pathsep.join([str(self._bin_dir), os.environ["PATH"]])

    def unset_env(self, key):
        if key in os.environ:
            del os.environ[key]

    def get_shell(self):
        shell = Path(os.environ.get("SHELL", "")).stem
        if shell in ("bash", "zsh", "fish"):
            return shell

    def _bin(self, bin):  # type: (str) -> str
        """
        Return path to the given executable.
        """
        if not self.is_venv():
            return bin

        bin_path = (self._bin_dir / bin).with_suffix(".exe" if self._windows else "")
        if not bin_path.exists():
            return bin

        return str(bin_path)

    def is_venv(self):  # type: () -> bool
        return self._venv is not None


class NullVenv(Venv):
    def __init__(self, execute=False):
        super(NullVenv, self).__init__()

        self.executed = []
        self._execute = execute

    def run(self, bin, *args):
        self.executed.append([bin] + list(args))

        if self._execute:
            return super(NullVenv, self).run(bin, *args)

    def _bin(self, bin):
        return bin



import poetry.packages

from poetry.semver import parse_constraint
from poetry.semver import Version
from poetry.semver import VersionConstraint
from poetry.semver import VersionRange
from poetry.semver import VersionUnion
from poetry.utils.helpers import canonicalize_name

from .constraints.empty_constraint import EmptyConstraint
from .constraints.generic_constraint import GenericConstraint
from .constraints.multi_constraint import MultiConstraint


class Dependency(object):
    def __init__(
        self,
        name,  # type: str
        constraint,  # type: str
        optional=False,  # type: bool
        category="main",  # type: str
        allows_prereleases=False,  # type: bool
    ):
        self._name = canonicalize_name(name)
        self._pretty_name = name

        try:
            if not isinstance(constraint, VersionConstraint):
                self._constraint = parse_constraint(constraint)
            else:
                self._constraint = constraint
        except ValueError:
            self._constraint = parse_constraint("*")

        self._pretty_constraint = str(constraint)
        self._optional = optional
        self._category = category

        if isinstance(self._constraint, VersionRange) and self._constraint.min:
            allows_prereleases = (
                allows_prereleases or self._constraint.min.is_prerelease()
            )

        self._allows_prereleases = allows_prereleases

        self._python_versions = "*"
        self._python_constraint = parse_constraint("*")
        self._platform = "*"
        self._platform_constraint = EmptyConstraint()

        self._extras = []
        self._in_extras = []

        self._activated = not self._optional

        self.is_root = False

    @property
    def name(self):
        return self._name

    @property
    def constraint(self):
        return self._constraint

    @property
    def pretty_constraint(self):
        return self._pretty_constraint

    @property
    def pretty_name(self):
        return self._pretty_name

    @property
    def category(self):
        return self._category

    @property
    def python_versions(self):
        return self._python_versions

    @python_versions.setter
    def python_versions(self, value):
        self._python_versions = value
        self._python_constraint = parse_constraint(value)

    @property
    def python_constraint(self):
        return self._python_constraint

    @property
    def platform(self):
        return self._platform

    @platform.setter
    def platform(self, value):
        self._platform = value
        self._platform_constraint = GenericConstraint.parse(value)

    @property
    def platform_constraint(self):
        return self._platform_constraint

    @property
    def extras(self):  # type: () -> list
        return self._extras

    @property
    def in_extras(self):  # type: () -> list
        return self._in_extras

    def allows_prereleases(self):
        return self._allows_prereleases

    def is_optional(self):
        return self._optional

    def is_activated(self):
        return self._activated

    def is_vcs(self):
        return False

    def is_file(self):
        return False

    def is_directory(self):
        return False

    def accepts(self, package):  # type: (poetry.packages.Package) -> bool
        """
        Determines if the given package matches this dependency.
        """
        return (
            self._name == package.name
            and self._constraint.allows(package.version)
            and (not package.is_prerelease() or self.allows_prereleases())
        )

    def to_pep_508(self, with_extras=True):  # type: (bool) -> str
        requirement = self.pretty_name

        if self.extras:
            requirement += "[{}]".format(",".join(self.extras))

        if isinstance(self.constraint, VersionUnion):
            requirement += " ({})".format(
                ",".join([str(c).replace(" ", "") for c in self.constraint.ranges])
            )
        elif isinstance(self.constraint, Version):
            requirement += " (=={})".format(self.constraint.text)
        elif not self.constraint.is_any():
            requirement += " ({})".format(str(self.constraint).replace(" ", ""))

        # Markers
        markers = []

        # Python marker
        if self.python_versions != "*":
            python_constraint = self.python_constraint

            markers.append(
                self._create_nested_marker("python_version", python_constraint)
            )

        if self.platform != "*":
            platform_constraint = self.platform_constraint

            markers.append(
                self._create_nested_marker("sys_platform", platform_constraint)
            )

        in_extras = " || ".join(self._in_extras)
        if in_extras and with_extras:
            markers.append(
                self._create_nested_marker("extra", GenericConstraint.parse(in_extras))
            )

        if markers:
            if len(markers) > 1:
                markers = ["({})".format(m) for m in markers]
                requirement += "; {}".format(" and ".join(markers))
            else:
                requirement += "; {}".format(markers[0])

        return requirement

    def _create_nested_marker(self, name, constraint):
        if isinstance(constraint, MultiConstraint):
            parts = []
            for c in constraint.constraints:
                multi = False
                if isinstance(c, MultiConstraint):
                    multi = True

                parts.append((multi, self._create_nested_marker(name, c)))

            glue = " and "
            if constraint.is_disjunctive():
                parts = [
                    "({})".format(part[1]) if part[0] else part[1] for part in parts
                ]
                glue = " or "
            else:
                parts = [part[1] for part in parts]

            marker = glue.join(parts)
        elif isinstance(constraint, GenericConstraint):
            marker = '{} {} "{}"'.format(
                name, constraint.string_operator, constraint.version
            )
        elif isinstance(constraint, VersionUnion):
            parts = []
            for c in constraint.ranges:
                parts.append(self._create_nested_marker(name, c))

            glue = " or "
            parts = ["({})".format(part) for part in parts]

            marker = glue.join(parts)
        elif isinstance(constraint, Version):
            marker = '{} == "{}"'.format(name, constraint.text)
        else:
            if constraint.min is not None:
                op = ">="
                if not constraint.include_min:
                    op = ">"

                version = constraint.min.text
                if constraint.max is not None:
                    text = '{} {} "{}"'.format(name, op, version)

                    op = "<="
                    if not constraint.include_max:
                        op = "<"

                    version = constraint.max

                    text += ' and {} {} "{}"'.format(name, op, version)

                    return text
            elif constraint.max is not None:
                op = "<="
                if not constraint.include_max:
                    op = "<"

                version = constraint.max
            else:
                return ""

            marker = '{} {} "{}"'.format(name, op, version)

        return marker

    def activate(self):
        """
        Set the dependency as mandatory.
        """
        self._activated = True

    def deactivate(self):
        """
        Set the dependency as optional.
        """
        if not self._optional:
            self._optional = True

        self._activated = False

    def with_constraint(self, constraint):
        new = Dependency(
            self.pretty_name,
            constraint,
            optional=self.is_optional(),
            category=self.category,
            allows_prereleases=self.allows_prereleases(),
        )

        new.is_root = self.is_root
        new.python_versions = self.python_versions
        new.platform = self.platform

        for extra in self.extras:
            new.extras.append(extra)

        for in_extra in self.in_extras:
            new.in_extras.append(in_extra)

        return new

    def __eq__(self, other):
        if not isinstance(other, Dependency):
            return NotImplemented

        return self._name == other.name and self._constraint == other.constraint

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((self._name, self._pretty_constraint))

    def __str__(self):
        if self.is_root:
            return self._pretty_name

        return "{} ({})".format(self._pretty_name, self._pretty_constraint)

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, str(self))



import codecs
import functools
import operator
import re
import string

from . import (
    TYPE_BOOLEAN,
    TYPE_INTEGER,
    TYPE_FLOAT,
    TYPE_DATE,
    TYPE_MULTILINE_STRING,
    TYPE_BARE_STRING,
    TYPE_MULTILINE_LITERAL_STRING,
    TYPE_LITERAL_STRING,
    TYPE_STRING,
)

from .errors import MalformedDateError
from .errors import BadEscapeCharacter


def deserialize(token):
    """
    Deserializes the value of a single tokens.Token instance based on its type.

    Raises DeserializationError when appropriate.
    """

    if token.type == TYPE_BOOLEAN:
        return _to_boolean(token)
    elif token.type == TYPE_INTEGER:
        return _to_int(token)
    elif token.type == TYPE_FLOAT:
        return _to_float(token)
    elif token.type == TYPE_DATE:
        return _to_date(token)
    elif token.type in (
        TYPE_STRING,
        TYPE_MULTILINE_STRING,
        TYPE_BARE_STRING,
        TYPE_LITERAL_STRING,
        TYPE_MULTILINE_LITERAL_STRING,
    ):
        return _to_string(token)
    else:
        raise Exception("This should never happen!")


def _unescape_str(text):
    """
    Unescapes a string according the TOML spec. Raises BadEscapeCharacter when appropriate.
    """

    # Detect bad escape jobs
    bad_escape_regexp = re.compile(r'([^\\]|^)\\[^btnfr"\\uU]')
    if bad_escape_regexp.findall(text):
        raise BadEscapeCharacter

    # Do the unescaping
    return codecs.decode(_unicode_escaped_string(text), "unicode-escape")


def _unicode_escaped_string(text):
    """
    Escapes all unicode characters in the given string
    """

    def is_unicode(c):
        return (
            c.lower()
            not in string.ascii_letters
            + string.whitespace
            + string.punctuation
            + string.digits
        )

    def escape_unicode_char(x):
        return codecs.encode(x, "unicode-escape")

    if any(is_unicode(c) for c in text):
        homogeneous_chars = tuple(
            escape_unicode_char(c) if is_unicode(c) else c.encode() for c in text
        )
        homogeneous_bytes = functools.reduce(operator.add, homogeneous_chars)
        return homogeneous_bytes.decode()
    else:
        return text


def _to_string(token):
    if token.type == TYPE_BARE_STRING:
        return token.source_substring

    elif token.type == TYPE_STRING:
        escaped = token.source_substring[1:-1]
        return _unescape_str(escaped)

    elif token.type == TYPE_MULTILINE_STRING:
        escaped = token.source_substring[3:-3]

        # Drop the first newline if existed
        if escaped and escaped[0] == "\n":
            escaped = escaped[1:]

        # Remove all occurrences of a slash-newline-zero-or-more-whitespace patterns
        escaped = re.sub(r"\\\n\s*", repl="", string=escaped, flags=re.DOTALL)
        return _unescape_str(escaped)

    elif token.type == TYPE_LITERAL_STRING:
        return token.source_substring[1:-1]

    elif token.type == TYPE_MULTILINE_LITERAL_STRING:
        text = token.source_substring[3:-3]
        if text[0] == "\n":
            text = text[1:]
        return text

    raise RuntimeError("Control should never reach here.")


def _to_int(token):
    return int(token.source_substring.replace("_", ""))


def _to_float(token):
    assert token.type == TYPE_FLOAT
    string = token.source_substring.replace("_", "")

    return float(string)


def _to_boolean(token):
    return token.source_substring == "true"


_correct_date_format = re.compile(
    r"(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})(Z|([+-])(\d{2}):(\d{2}))"
)


def _to_date(token):
    if not _correct_date_format.match(token.source_substring):
        raise MalformedDateError

    return token.source_substring



import os
import posixpath
import re

try:
    import urllib.parse as urlparse
except ImportError:
    import urlparse


try:
    import urllib.request as urllib2
except ImportError:
    import urllib2


BZ2_EXTENSIONS = (".tar.bz2", ".tbz")
XZ_EXTENSIONS = (".tar.xz", ".txz", ".tlz", ".tar.lz", ".tar.lzma")
ZIP_EXTENSIONS = (".zip", ".whl")
TAR_EXTENSIONS = (".tar.gz", ".tgz", ".tar")
ARCHIVE_EXTENSIONS = ZIP_EXTENSIONS + BZ2_EXTENSIONS + TAR_EXTENSIONS + XZ_EXTENSIONS
SUPPORTED_EXTENSIONS = ZIP_EXTENSIONS + TAR_EXTENSIONS

try:
    import bz2  # noqa

    SUPPORTED_EXTENSIONS += BZ2_EXTENSIONS
except ImportError:
    pass

try:
    # Only for Python 3.3+
    import lzma  # noqa

    SUPPORTED_EXTENSIONS += XZ_EXTENSIONS
except ImportError:
    pass


def path_to_url(path):
    """
    Convert a path to a file: URL.  The path will be made absolute and have
    quoted path parts.
    """
    path = os.path.normpath(os.path.abspath(path))
    url = urlparse.urljoin("file:", urllib2.pathname2url(path))
    return url


def is_url(name):
    if ":" not in name:
        return False
    scheme = name.split(":", 1)[0].lower()

    return scheme in [
        "http",
        "https",
        "file",
        "ftp",
        "ssh",
        "git",
        "hg",
        "bzr",
        "sftp",
        "svn" "ssh",
    ]


def strip_extras(path):
    m = re.match(r"^(.+)(\[[^\]]+\])$", path)
    extras = None
    if m:
        path_no_extras = m.group(1)
        extras = m.group(2)
    else:
        path_no_extras = path

    return path_no_extras, extras


def is_installable_dir(path):
    """Return True if `path` is a directory containing a setup.py file."""
    if not os.path.isdir(path):
        return False
    setup_py = os.path.join(path, "setup.py")
    if os.path.isfile(setup_py):
        return True
    return False


def is_archive_file(name):
    """Return True if `name` is a considered as an archive file."""
    ext = splitext(name)[1].lower()
    if ext in ARCHIVE_EXTENSIONS:
        return True
    return False


def splitext(path):
    """Like os.path.splitext, but take off .tar too"""
    base, ext = posixpath.splitext(path)
    if base.lower().endswith(".tar"):
        ext = base[-4:] + ext
        base = base[:-4]
    return base, ext


def group_markers(markers):
    groups = [[]]

    for marker in markers:
        assert isinstance(marker, (list, tuple, str))

        if isinstance(marker, list):
            groups[-1].append(group_markers(marker))
        elif isinstance(marker, tuple):
            lhs, op, rhs = marker

            groups[-1].append((lhs.value, op, rhs.value))
        else:
            assert marker in ["and", "or"]
            if marker == "or":
                groups.append([])

    return groups


def convert_markers(markers):
    groups = group_markers(markers)

    requirements = {}

    def _group(_groups, or_=False):
        for group in _groups:
            if isinstance(group, tuple):
                variable, op, value = group
                group_name = str(variable)
                if group_name not in requirements:
                    requirements[group_name] = [[]]
                elif or_:
                    requirements[group_name].append([])

                or_ = False

                requirements[group_name][-1].append((str(op), str(value)))
            else:
                _group(group, or_=True)

    _group(groups)

    return requirements



import re

from .constraint import Constraint


class WilcardConstraint(Constraint):
    def __init__(self, constraint):  # type: (str) -> None
        m = re.match(
            "^(!= ?|==)?v?(\d+)(?:\.(\d+))?(?:\.(\d+))?(?:\.[xX*])+$", constraint
        )
        if not m:
            raise ValueError("Invalid value for wildcard constraint")

        if not m.group(1):
            operator = "=="
        else:
            operator = m.group(1).strip()

        super(WilcardConstraint, self).__init__(
            operator, ".".join([g if g else "*" for g in m.groups()[1:]])
        )

        if m.group(4):
            position = 2
        elif m.group(3):
            position = 1
        else:
            position = 0

        from ..version_parser import VersionParser

        parser = VersionParser()
        groups = m.groups()[1:]
        low_version = parser._manipulate_version_string(groups, position)
        high_version = parser._manipulate_version_string(groups, position, 1)

        if operator == "!=":
            if low_version == "0.0.0.0":
                self._constraint = Constraint(">=", high_version)
            else:
                self._constraint = parser.parse_constraints(
                    "<{} || >={}".format(low_version, high_version)
                )
        else:
            if low_version == "0.0.0.0":
                self._constraint = Constraint("<", high_version)
            else:
                self._constraint = parser.parse_constraints(
                    ">={},<{}".format(low_version, high_version)
                )

    @property
    def supported_operators(self):
        return ["!=", "=="]

    @property
    def constraint(self):
        return self._constraint

    def matches(self, provider):  # type: (Constraint) -> bool
        if isinstance(provider, self.__class__):
            return self._constraint.matches(provider.constraint)

        return provider.matches(self._constraint)

    def __str__(self):
        op = ""
        if self.string_operator == "!=":
            op = "!= "

        return "{}{}".format(op, self._version)



import os

from .venv_command import VenvCommand


class DevelopCommand(VenvCommand):
    """
    Installs the current project in development mode.

    develop
    """

    help = """\
The <info>develop</info> command installs the current project in development mode.
"""

    def handle(self):
        from poetry.masonry.builders import SdistBuilder
        from poetry.io import NullIO
        from poetry.utils._compat import decode
        from poetry.utils.venv import NullVenv

        setup = self.poetry.file.parent / "setup.py"
        has_setup = setup.exists()

        if has_setup:
            self.line("<warning>A setup.py file already exists. Using it.</warning>")
        else:
            builder = SdistBuilder(self.poetry, NullVenv(), NullIO())

            with setup.open("w") as f:
                f.write(decode(builder.build_setup()))

        try:
            self._install(setup)
        finally:
            if not has_setup:
                os.remove(str(setup))

    def _install(self, setup):
        self.call("install")

        self.line(
            "Installing <info>{}</info> (<comment>{}</comment>)".format(
                self.poetry.package.pretty_name, self.poetry.package.pretty_version
            )
        )
        self.venv.run("pip", "install", "-e", str(setup.parent), "--no-deps")



from .command import Command


class CheckCommand(Command):
    """
    Checks the validity of the <comment>pyproject.toml</comment> file.

    check
    """

    def handle(self):
        # Load poetry and display errors, if any
        self.poetry.check(self.poetry.local_config, strict=True)

        self.info("All set!")



# -*- coding: utf-8 -*-
import pytoml as toml

from poetry.toml import dumps
from poetry.toml import loads
from poetry.toml import TOMLFile

from ._compat import Path


class TomlFile:
    def __init__(self, path):
        self._path = Path(path)

    @property
    def path(self):
        return self._path

    def read(self, raw=False):  # type: (bool) -> dict
        with self._path.open(encoding="utf-8") as f:
            if raw:
                return toml.loads(f.read())

            return loads(f.read())

    def dumps(self, data, sort=False):  # type: (...) -> str
        if not isinstance(data, TOMLFile):
            data = toml.dumps(data, sort_keys=sort)
        else:
            data = dumps(data)

        return data

    def write(self, data, sort=False):  # type: (...) -> None
        data = self.dumps(data, sort=sort)

        with self._path.open("w", encoding="utf-8") as f:
            f.write(data)

    def __getattr__(self, item):
        return getattr(self._path, item)
